import base64
import os.path
import random
import threading
from http.server import BaseHTTPRequestHandler, HTTPServer
from pathlib import Path
from typing import Tuple
from urllib.parse import parse_qs, urlparse

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import padding, rsa
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPrivateKey, RSAPublicKey
from rich.console import Console

from recce.event import update_recce_api_token
from recce.exceptions import RecceConfigException
from recce.util.onboarding_state import update_onboarding_state
from recce.util.recce_cloud import RECCE_CLOUD_BASE_URL, RecceCloud

console = Console()

static_folder_path = Path(__file__).parent / "data"
_server_lock = threading.Lock()
_connection_url = None


def decrypt_code(private_key: RSAPrivateKey, code: str) -> str:
    ciphertext = base64.b64decode(code)
    plaintext = private_key.decrypt(
        ciphertext,
        padding.OAEP(
            mgf=padding.MGF1(algorithm=hashes.SHA1()),  # Node.js uses SHA1 by default
            algorithm=hashes.SHA1(),
            label=None,
        ),
    )
    return plaintext.decode("utf-8")


def handle_callback_request(query_string: str, private_key: RSAPrivateKey):
    query_params = parse_qs(query_string)
    code = query_params.get("code", [None])[0]
    if not code:
        raise RecceConfigException("Missing `code` in query")

    api_token = decrypt_code(private_key, code)
    if not RecceCloud(api_token).verify_token():
        raise RecceConfigException("Invalid Recce Cloud API token")

    update_recce_api_token(api_token)
    update_onboarding_state(api_token, False)

    return api_token  # for testability/debugging


def make_callback_handler(private_key: RSAPrivateKey):
    class OneTimeHTTPRequestHandler(BaseHTTPRequestHandler):
        def do_GET(self):
            try:
                with open(os.path.join(static_folder_path, "auth_callback.html"), "r", encoding="utf-8") as f:
                    callback_html_content = f.read()

                # Parse query parameters
                parsed_url = urlparse(self.path)

                handle_callback_request(parsed_url.query, private_key)

                # Construct HTML content
                self.send_response(200)
                self.send_header("Content-Type", "text/html")
                self.send_header("Content-Length", str(len(callback_html_content.encode())))
                self.end_headers()
                self.wfile.write(callback_html_content.encode())

            except Exception:
                console.print_exception()
                self.send_response(500)
                self.end_headers()
                self.wfile.write(b"<h1>Internal Server Error</h1>")
            finally:
                # Shut down the server after handling the first request
                # Shutdown in a new thread to avoid deadlock
                self.server.server_close()
                threading.Thread(target=self.server.shutdown, daemon=True).start()

        def log_message(self, format, *args):
            # Suppress default logging
            return

    return OneTimeHTTPRequestHandler


def is_callback_server_running():
    return _server_lock.locked()


def get_connection_url():
    return _connection_url


def run_one_time_http_server(private_key: RSAPrivateKey, port=8080):
    handler = make_callback_handler(private_key)
    server = HTTPServer(("localhost", port), handler)
    server.serve_forever()


def prepare_connection_url(public_key: RSAPublicKey):
    public_key_pem_bytes = public_key.public_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PublicFormat.SubjectPublicKeyInfo,
    )
    public_key_pem_str = base64.b64encode(public_key_pem_bytes).decode("utf-8")
    callback_port = random.randint(10000, 15000)
    connect_url = f"{RECCE_CLOUD_BASE_URL}/connect?_key={public_key_pem_str}&_port={callback_port}"
    return connect_url, callback_port


def generate_key_pair() -> Tuple[RSAPrivateKey, RSAPublicKey]:
    key_size = 2048  # Should be at least 2048

    private_key = rsa.generate_private_key(
        public_exponent=65537, key_size=key_size, backend=default_backend()  # Do not change
    )

    public_key = private_key.public_key()
    return private_key, public_key


def connect_to_cloud_background_task(private_key: RSAPrivateKey, callback_port, connection_url):
    if is_callback_server_running():
        return

    with _server_lock:
        global _connection_url
        _connection_url = connection_url
        run_one_time_http_server(private_key, callback_port)
        _connection_url = None



from unittest.mock import MagicMock, patch

import pytest

# Skip all tests in this module if mcp is not available
pytest.importorskip("mcp")

from recce.core import RecceContext  # noqa: E402
from recce.mcp_server import RecceMCPServer, run_mcp_server  # noqa: E402
from recce.models.types import LineageDiff  # noqa: E402
from recce.tasks.profile import ProfileDiffTask  # noqa: E402
from recce.tasks.query import QueryDiffTask, QueryTask  # noqa: E402
from recce.tasks.rowcount import RowCountDiffTask  # noqa: E402


@pytest.fixture
def mcp_server():
    """Fixture to create a RecceMCPServer instance for testing"""
    mock_context = MagicMock(spec=RecceContext)
    return RecceMCPServer(mock_context), mock_context


class TestRecceMCPServer:
    """Test cases for the RecceMCPServer class"""

    def test_server_initialization(self, mcp_server):
        """Test that the MCP server initializes correctly"""
        server, mock_context = mcp_server
        assert server.context == mock_context
        assert server.server is not None
        assert server.server.name == "recce"

    @pytest.mark.asyncio
    async def test_get_lineage_diff(self, mcp_server):
        """Test the get_lineage_diff tool"""
        server, mock_context = mcp_server
        # Mock the lineage diff response
        mock_lineage_diff = MagicMock(spec=LineageDiff)
        mock_lineage_diff.model_dump.return_value = {
            "added": ["model.project.new_model"],
            "removed": ["model.project.old_model"],
            "modified": ["model.project.changed_model"],
        }
        mock_context.get_lineage_diff.return_value = mock_lineage_diff

        # Execute the method
        result = await server._get_lineage_diff({})

        # Verify the result
        assert "added" in result
        assert "removed" in result
        assert "modified" in result
        mock_context.get_lineage_diff.assert_called_once()
        mock_lineage_diff.model_dump.assert_called_once_with(mode="json")

    @pytest.mark.asyncio
    async def test_row_count_diff(self, mcp_server):
        """Test the row_count_diff tool"""
        server, _ = mcp_server
        # Mock the task execution
        mock_result = {"results": [{"node_id": "model.project.my_model", "base": 100, "current": 105, "diff": 5}]}

        with patch.object(RowCountDiffTask, "execute", return_value=mock_result):
            result = await server._row_count_diff({"node_names": ["my_model"]})

        # Verify the result
        assert result == mock_result
        assert "results" in result

    @pytest.mark.asyncio
    async def test_query(self, mcp_server):
        """Test the query tool"""
        server, _ = mcp_server
        # Mock the task execution
        mock_result = MagicMock()
        mock_result.model_dump.return_value = {
            "columns": ["id", "name"],
            "data": [[1, "Alice"], [2, "Bob"]],
        }

        with patch.object(QueryTask, "execute", return_value=mock_result):
            result = await server._query({"sql_template": "SELECT * FROM {{ ref('my_model') }}", "base": False})

        # Verify the result
        assert "columns" in result
        assert "data" in result
        mock_result.model_dump.assert_called_once_with(mode="json")

    @pytest.mark.asyncio
    async def test_query_with_base_flag(self, mcp_server):
        """Test the query tool with base environment flag"""
        server, _ = mcp_server
        mock_result = {"columns": ["id"], "data": [[1]]}

        with patch.object(QueryTask, "execute", return_value=mock_result) as mock_execute:
            with patch.object(QueryTask, "__init__", return_value=None):
                task = QueryTask(params={"sql_template": "SELECT 1"})
                task.is_base = True
                task.execute = mock_execute

                result = await server._query({"sql_template": "SELECT 1", "base": True})

                # Verify base flag was set (would need to inspect task creation)
                assert result == mock_result

    @pytest.mark.asyncio
    async def test_query_diff(self, mcp_server):
        """Test the query_diff tool"""
        server, _ = mcp_server
        # Mock the task execution
        mock_result = MagicMock()
        mock_result.model_dump.return_value = {
            "diff": {
                "added": [[3, "Charlie"]],
                "removed": [[1, "Alice"]],
                "modified": [],
            }
        }

        with patch.object(QueryDiffTask, "execute", return_value=mock_result):
            result = await server._query_diff(
                {
                    "sql_template": "SELECT * FROM {{ ref('my_model') }}",
                    "primary_keys": ["id"],
                }
            )

        # Verify the result
        assert "diff" in result
        mock_result.model_dump.assert_called_once_with(mode="json")

    @pytest.mark.asyncio
    async def test_profile_diff(self, mcp_server):
        """Test the profile_diff tool"""
        server, _ = mcp_server
        # Mock the task execution
        mock_result = MagicMock()
        mock_result.model_dump.return_value = {
            "columns": {
                "id": {
                    "base": {"min": 1, "max": 100, "avg": 50.5},
                    "current": {"min": 1, "max": 105, "avg": 53.0},
                }
            }
        }

        with patch.object(ProfileDiffTask, "execute", return_value=mock_result):
            result = await server._profile_diff({"model": "my_model", "columns": ["id"]})

        # Verify the result
        assert "columns" in result
        mock_result.model_dump.assert_called_once_with(mode="json")

    @pytest.mark.asyncio
    async def test_error_handling(self, mcp_server):
        """Test error handling in tool execution"""
        server, mock_context = mcp_server
        # Make get_lineage_diff raise an exception
        mock_context.get_lineage_diff.side_effect = Exception("Test error")

        # The method should raise the exception
        with pytest.raises(Exception, match="Test error"):
            await server._get_lineage_diff({})


class TestRunMCPServer:
    """Test cases for the run_mcp_server function"""

    @pytest.mark.asyncio
    @patch("recce.mcp_server.load_context")
    @patch.object(RecceMCPServer, "run")
    async def test_run_mcp_server(self, mock_run, mock_load_context):
        """Test the run_mcp_server entry point"""
        # Mock the context
        mock_context = MagicMock(spec=RecceContext)
        mock_load_context.return_value = mock_context

        # Mock the server run method
        mock_run.return_value = None

        # Run the server
        await run_mcp_server(project_dir="/test/path")

        # Verify context was loaded with correct kwargs
        mock_load_context.assert_called_once_with(project_dir="/test/path")

        # Verify server was run
        mock_run.assert_called_once()

    @pytest.mark.asyncio
    @patch("recce.mcp_server.load_context")
    async def test_run_mcp_server_context_error(self, mock_load_context):
        """Test run_mcp_server handles context loading errors"""
        # Make load_context raise an exception
        mock_load_context.side_effect = FileNotFoundError("manifest.json not found")

        # The function should raise the exception
        with pytest.raises(FileNotFoundError):
            await run_mcp_server()


def test_mcp_cli_command_exists():
    """Test that the mcp-server CLI command is registered"""
    from recce.cli import cli

    # Check that mcp_server is in the CLI commands
    commands = [cmd.name for cmd in cli.commands.values()]
    assert "mcp_server" in commands or "mcp-server" in commands



import json
from unittest.mock import patch

import pytest
import requests

from recce.pull_request import _fetch_pr_title, fetch_pr_metadata_from_event_path


@pytest.fixture
def mock_github_event(tmp_path, monkeypatch):
    event_data = {
        "number": 1,
        "pull_request": {
            "_links": {
                "html": {"href": "https://github.com/xyz/abc/pull/1"},
                "self": {"href": "api_url_here"},
            }
        },
    }
    event_file = tmp_path / "github_event.json"
    with open(event_file, "w") as f:
        json.dump(event_data, f)

    monkeypatch.setenv("GITHUB_EVENT_PATH", str(event_file))
    monkeypatch.setenv("GITHUB_REPOSITORY", "abc/xyz")


@pytest.fixture
def mock_github_token(monkeypatch):
    monkeypatch.setenv("GITHUB_TOKEN", "github_token_here")


class MockResponse:
    def __init__(self, status_code, json_data=None):
        self.status_code = status_code
        self.json_data = json_data

    def json(self):
        return self.json_data


@pytest.fixture
def mock_get_request_success(monkeypatch):
    def mock_get(*args, **kwargs):
        return MockResponse(200, json_data={"title": "Update README.md"})

    monkeypatch.setattr(requests, "get", mock_get)


@pytest.fixture
def mock_get_request_failure(monkeypatch):
    def mock_get(*args, **kwargs):
        return MockResponse(404)

    monkeypatch.setattr(requests, "get", mock_get)


def test_fetch_pr_metadata(mock_github_event, mock_github_token, mock_get_request_success):
    result = fetch_pr_metadata_from_event_path()

    assert result is not None
    assert result["github_pr_id"] == 1
    assert result["github_pr_url"] == "https://github.com/xyz/abc/pull/1"
    assert result["github_pr_title"] == "Update README.md"
    assert result["github_repository"] == "abc/xyz"


def test_fetch_pr_metadata_no_event(mock_get_request_success, monkeypatch):
    monkeypatch.delenv("GITHUB_EVENT_PATH", raising=False)

    result = fetch_pr_metadata_from_event_path()

    assert result is None


def test_fetch_pr_metadata_request_error(mock_github_event, mock_get_request_failure):
    result = fetch_pr_metadata_from_event_path()

    assert result is not None
    assert result["github_pr_id"] == 1
    assert result["github_pr_url"] == "https://github.com/xyz/abc/pull/1"
    assert result["github_pr_title"] is None


def test_fetch_pr_metadata_exception(mock_github_event, monkeypatch):
    monkeypatch.setenv("GITHUB_EVENT_PATH", "nonexistent_path")

    result = fetch_pr_metadata_from_event_path()

    assert result is None


def test_fetch_pr_metadata_no_token(mock_github_event):
    with patch.dict("os.environ", clear=True):
        result = fetch_pr_metadata_from_event_path()

    assert result is None


def test_fetch_pr_title(mock_github_token, mock_get_request_success):
    result = _fetch_pr_title("api_url_here")

    assert result == "Update README.md"


def test_fetch_pr_title_request_error(mock_github_token, mock_get_request_failure):
    result = _fetch_pr_title("api_url_here")

    assert result is None


def test_fetch_pr_title_exception(mock_github_token, monkeypatch):
    monkeypatch.setenv("GITHUB_TOKEN", "github_token_here")

    def mock_get(*args, **kwargs):
        raise Exception("Test exception")

    monkeypatch.setattr(requests, "get", mock_get)

    result = _fetch_pr_title("api_url_here")

    assert result is None


def test_fetch_pr_title_no_token():
    with patch.dict("os.environ", clear=True):
        result = _fetch_pr_title("api_url_here")

    assert result is None



import json
import logging
import os
from typing import IO, Dict

import requests

from recce import get_version
from recce.event import get_user_id, is_anonymous_tracking
from recce.pull_request import PullRequestInfo

RECCE_CLOUD_API_HOST = os.environ.get("RECCE_CLOUD_API_HOST", "https://cloud.datarecce.io")
RECCE_CLOUD_BASE_URL = os.environ.get("RECCE_CLOUD_BASE_URL", RECCE_CLOUD_API_HOST)

DOCKER_INTERNAL_URL_PREFIX = "http://host.docker.internal"
LOCALHOST_URL_PREFIX = "http://localhost"

logger = logging.getLogger("uvicorn")


class PresignedUrlMethod:
    UPLOAD = "upload"
    DOWNLOAD = "download"


class RecceCloudException(Exception):
    def __init__(self, message: str, reason: str, status_code: int):
        super().__init__(message)
        self.status_code = status_code

        try:
            reason = json.loads(reason).get("detail", "")
        except json.JSONDecodeError:
            pass
        self.reason = reason


class RecceCloud:
    def __init__(self, token: str):
        if token is None:
            raise ValueError("Token cannot be None.")
        self.token = token
        self.token_type = "github_token" if token.startswith(("ghp_", "gho_", "ghu_", "ghs_", "ghr_")) else "api_token"
        self.base_url = f"{RECCE_CLOUD_API_HOST}/api/v1"
        self.base_url_v2 = f"{RECCE_CLOUD_API_HOST}/api/v2"

    def _request(self, method, url, headers: Dict = None, **kwargs):
        headers = {
            **(headers or {}),
            "Authorization": f"Bearer {self.token}",
        }
        return requests.request(method, url, headers=headers, **kwargs)

    def verify_token(self) -> bool:
        if self.token_type == "github_token":
            return True
        # Verify the Recce Cloud API token
        api_url = f"{self.base_url}/verify-token"
        try:
            headers: Dict = None
            if is_anonymous_tracking():
                headers = {
                    "X-Recce-Oss-User-Id": get_user_id(),
                    "X-Recce-Oss-Version": get_version(),
                }
            response = self._request("GET", api_url, headers=headers)
            if response.status_code == 200:
                return True
        except Exception:
            pass
        return False

    def get_presigned_url_by_github_repo(
        self,
        method: PresignedUrlMethod,
        repository: str,
        artifact_name: str,
        metadata: dict = None,
        pr_id: int = None,
        branch: str = None,
    ) -> str:
        response = self._fetch_presigned_url(method, repository, artifact_name, metadata, pr_id, branch)
        return response.get("presigned_url")

    def _replace_localhost_with_docker_internal(self, url: str) -> str:
        if url is None:
            return None
        if (
            os.environ.get("RECCE_SHARE_INSTANCE_ENV") == "docker"
            or os.environ.get("RECCE_TASK_INSTANCE_ENV") == "docker"
            or os.environ.get("RECCE_INSTANCE_ENV") == "docker"
        ):
            # For local development, convert the presigned URL from localhost to host.docker.internal
            if url.startswith(LOCALHOST_URL_PREFIX):
                return url.replace(LOCALHOST_URL_PREFIX, DOCKER_INTERNAL_URL_PREFIX)
        return url

    def get_presigned_url_by_share_id(
        self,
        method: PresignedUrlMethod,
        share_id: str,
        metadata: dict = None,
    ) -> str:
        response = self._fetch_presigned_url_by_share_id(method, share_id, metadata=metadata)
        presigned_url = response.get("presigned_url")
        if not presigned_url:
            raise RecceCloudException(
                message="Failed to get presigned URL from Recce Cloud.",
                reason="No presigned URL returned from the server.",
                status_code=404,
            )
        presigned_url = self._replace_localhost_with_docker_internal(presigned_url)
        return presigned_url

    def get_download_presigned_url_by_github_repo_with_tags(
        self, repository: str, artifact_name: str, branch: str = None
    ) -> (str, dict):
        response = self._fetch_presigned_url(PresignedUrlMethod.DOWNLOAD, repository, artifact_name, branch=branch)
        return response.get("presigned_url"), response.get("tags", {})

    def _fetch_presigned_url(
        self,
        method: PresignedUrlMethod,
        repository: str,
        artifact_name: str,
        metadata: dict = None,
        pr_id: int = None,
        branch: str = None,
    ) -> str:
        if pr_id is not None:
            api_url = f"{self.base_url}/{repository}/pulls/{pr_id}/artifacts/{method}?artifact_name={artifact_name}&enable_ssec=true"
        elif branch is not None:
            api_url = f"{self.base_url}/{repository}/commits/{branch}/artifacts/{method}?artifact_name={artifact_name}&enable_ssec=true"
        else:
            raise ValueError("Either pr_id or sha must be provided.")
        response = self._request("POST", api_url, json=metadata)
        if response.status_code != 200:
            raise RecceCloudException(
                message="Failed to {method} artifact {preposition} Recce Cloud.".format(
                    method=method, preposition="from" if method == PresignedUrlMethod.DOWNLOAD else "to"
                ),
                reason=response.text,
                status_code=response.status_code,
            )
        return response.json()

    def _fetch_presigned_url_by_share_id(
        self,
        method: PresignedUrlMethod,
        share_id: str,
        metadata: dict = None,
    ):
        api_url = f"{self.base_url}/shares/{share_id}/presigned/{method}"
        data = None
        # Only provide metadata for upload requests
        if method == PresignedUrlMethod.UPLOAD:
            # Covert metadata values to strings to ensure JSON serializability
            data = {"metadata": {key: str(value) for key, value in metadata.items()}} if metadata else None
        response = self._request(
            "POST",
            api_url,
            json=data,
        )
        if response.status_code != 200:
            raise RecceCloudException(
                message="Failed to {method} artifact {preposition} Recce Cloud.".format(
                    method=method, preposition="from" if method == PresignedUrlMethod.DOWNLOAD else "to"
                ),
                reason=response.text,
                status_code=response.status_code,
            )
        return response.json()

    def get_artifact_metadata(self, pr_info: PullRequestInfo) -> dict:
        api_url = f"{self.base_url}/{pr_info.repository}/pulls/{pr_info.id}/metadata"
        response = self._request("GET", api_url)
        if response.status_code == 204:
            return None
        if response.status_code != 200:
            raise RecceCloudException(
                message="Failed to get artifact metadata from Recce Cloud.",
                reason=response.text,
                status_code=response.status_code,
            )
        return response.json()

    def purge_artifacts(self, repository: str, pr_id: int = None, branch: str = None):
        if pr_id is not None:
            api_url = f"{self.base_url}/{repository}/pulls/{pr_id}/artifacts"
            error_message = "Failed to purge artifacts from Recce Cloud."
        elif branch is not None:
            api_url = f"{self.base_url}/{repository}/commits/{branch}/artifacts"
            error_message = "Failed to delete artifacts from Recce Cloud."
        else:
            raise ValueError(
                "Please either run this command from within a pull request context "
                "or specify a branch using the --branch option."
            )
        response = self._request("DELETE", api_url)
        if response.status_code != 204:
            raise RecceCloudException(
                message=error_message,
                reason=response.text,
                status_code=response.status_code,
            )

    def check_artifacts_exists(self, pr_info: PullRequestInfo) -> bool:
        api_url = f"{self.base_url}/{pr_info.repository}/pulls/{pr_info.id}/metadata"
        response = self._request("GET", api_url)
        if response.status_code == 200:
            return True
        elif response.status_code == 204:
            return False
        else:
            raise RecceCloudException(
                message="Failed to check if artifacts exist in Recce Cloud.",
                reason=response.text,
                status_code=response.status_code,
            )

    def share_state(self, file_name: str, file_io: IO):
        api_url = f"{self.base_url}/recce-state/upload"
        files = {"file": (file_name, file_io, "application/json")}
        response = self._request("POST", api_url, files=files)
        if response.status_code == 403:
            return {"status": "error", "message": response.json().get("detail")}
        if response.status_code != 200:
            raise RecceCloudException(
                message="Failed to share Recce state.", reason=response.text, status_code=response.status_code
            )
        return response.json()

    def update_github_pull_request_check(self, pr_info: PullRequestInfo, metadata: dict = None):
        api_url = f"{self.base_url}/{pr_info.repository}/pulls/{pr_info.id}/github/checks"
        try:
            self._request("POST", api_url, json=metadata)
        except Exception as e:
            # We don't care the response of this request, so we don't need to raise any exception.
            logger.debug(f"Failed to update the GitHub PR check. Reason: {str(e)}")

    def get_user_info(self) -> Dict:
        api_url = f"{self.base_url}/users"
        response = self._request("GET", api_url)
        if response.status_code != 200:
            raise RecceCloudException(
                message="Failed to get user info from Recce Cloud.",
                reason=response.text,
                status_code=response.status_code,
            )
        return response.json().get("user")

    def set_onboarding_state(self, state: str):
        api_url = f"{self.base_url}/users/onboarding-state"
        try:
            response = self._request("PUT", api_url, json={"state": state})
            response.raise_for_status()
        except requests.exceptions.HTTPError as e:
            # Don't Raise an exception if setting onboarding_state fails
            logger.warning(f"Failed to set Onboarding State in Recce Cloud. Reason: {str(e)}")
        return

    def get_session(self, session_id: str):
        api_url = f"{self.base_url_v2}/sessions/{session_id}"
        response = self._request("GET", api_url)
        if response.status_code == 403:
            return {"status": "error", "message": response.json().get("detail")}
        if response.status_code != 200:
            raise RecceCloudException(
                message="Failed to get session from Recce Cloud.",
                reason=response.text,
                status_code=response.status_code,
            )
        data = response.json()
        if data["success"] is not True:
            raise RecceCloudException(
                message="Failed to get session from Recce Cloud.",
                reason=data.get("message", "Unknown error"),
                status_code=response.status_code,
            )
        return data["session"]

    def update_session(self, org_id: str, project_id: str, session_id: str, adapter_type: str):
        api_url = f"{self.base_url_v2}/organizations/{org_id}/projects/{project_id}/sessions/{session_id}"
        data = {"adapter_type": adapter_type}
        response = self._request("PATCH", api_url, json=data)
        if response.status_code == 403:
            return {"status": "error", "message": response.json().get("detail")}
        if response.status_code != 200:
            raise RecceCloudException(
                message="Failed to update session in Recce Cloud.",
                reason=response.text,
                status_code=response.status_code,
            )
        return response.json()

    def get_download_urls_by_session_id(self, org_id: str, project_id: str, session_id: str) -> dict[str, str]:
        api_url = f"{self.base_url_v2}/organizations/{org_id}/projects/{project_id}/sessions/{session_id}/download-url"
        response = self._request("GET", api_url)
        if response.status_code != 200:
            raise RecceCloudException(
                message="Failed to download session from Recce Cloud.",
                reason=response.text,
                status_code=response.status_code,
            )
        data = response.json()
        if data["presigned_urls"] is None:
            raise RecceCloudException(
                message="No presigned URLs returned from the server.",
                reason="",
                status_code=404,
            )

        presigned_urls = data["presigned_urls"]
        for key, url in presigned_urls.items():
            presigned_urls[key] = self._replace_localhost_with_docker_internal(url)
        return presigned_urls

    def get_base_session_download_urls(self, org_id: str, project_id: str) -> dict[str, str]:
        """Get download URLs for the base session of a project."""
        api_url = f"{self.base_url_v2}/organizations/{org_id}/projects/{project_id}/base-session/download-url"
        response = self._request("GET", api_url)
        if response.status_code != 200:
            raise RecceCloudException(
                message="Failed to download base session from Recce Cloud.",
                reason=response.text,
                status_code=response.status_code,
            )
        data = response.json()
        if data["presigned_urls"] is None:
            raise RecceCloudException(
                message="No presigned URLs returned from the server.",
                reason="",
                status_code=404,
            )

        presigned_urls = data["presigned_urls"]
        for key, url in presigned_urls.items():
            presigned_urls[key] = self._replace_localhost_with_docker_internal(url)
        return presigned_urls

    def get_upload_urls_by_session_id(self, org_id: str, project_id: str, session_id: str) -> dict[str, str]:
        api_url = f"{self.base_url_v2}/organizations/{org_id}/projects/{project_id}/sessions/{session_id}/upload-url"
        response = self._request("GET", api_url)
        if response.status_code != 200:
            raise RecceCloudException(
                message="Failed to get upload URLs for session from Recce Cloud.",
                reason=response.text,
                status_code=response.status_code,
            )
        data = response.json()
        if data["presigned_urls"] is None:
            raise RecceCloudException(
                message="No presigned URLs returned from the server.",
                reason="",
                status_code=404,
            )

        presigned_urls = data["presigned_urls"]
        for key, url in presigned_urls.items():
            presigned_urls[key] = self._replace_localhost_with_docker_internal(url)
        return presigned_urls

    def post_recce_state_uploaded_by_session_id(self, org_id: str, project_id: str, session_id: str):
        api_url = f"{self.base_url_v2}/organizations/{org_id}/projects/{project_id}/sessions/{session_id}/recce-state-uploaded"
        response = self._request("POST", api_url)
        if response.status_code != 204:
            raise RecceCloudException(
                message="Failed to notify state uploaded for session in Recce Cloud.",
                reason=response.text,
                status_code=response.status_code,
            )

    def list_organizations(self) -> list:
        """List all organizations the user has access to."""
        api_url = f"{self.base_url_v2}/organizations"
        response = self._request("GET", api_url)
        if response.status_code != 200:
            raise RecceCloudException(
                message="Failed to list organizations from Recce Cloud.",
                reason=response.text,
                status_code=response.status_code,
            )
        data = response.json()
        return data.get("organizations", [])

    def list_projects(self, org_id: str) -> list:
        """List all projects in an organization."""
        api_url = f"{self.base_url_v2}/organizations/{org_id}/projects"
        response = self._request("GET", api_url)
        if response.status_code != 200:
            raise RecceCloudException(
                message="Failed to list projects from Recce Cloud.",
                reason=response.text,
                status_code=response.status_code,
            )
        data = response.json()
        return data.get("projects", [])

    def list_sessions(self, org_id: str, project_id: str) -> list:
        """List all sessions in a project."""
        api_url = f"{self.base_url_v2}/organizations/{org_id}/projects/{project_id}/sessions"
        response = self._request("GET", api_url)
        if response.status_code != 200:
            raise RecceCloudException(
                message="Failed to list sessions from Recce Cloud.",
                reason=response.text,
                status_code=response.status_code,
            )
        data = response.json()
        return data.get("sessions", [])


def get_recce_cloud_onboarding_state(token: str) -> str:
    if token and token.startswith("rct-"):
        return "undefined"

    try:
        recce_cloud = RecceCloud(token)
        user_info = recce_cloud.get_user_info()
        if user_info:
            return user_info.get("onboarding_state")
    except Exception as e:
        logger.debug(str(e))
    return "undefined"


def set_recce_cloud_onboarding_state(token: str, new_state: str):
    recce_cloud = RecceCloud(token)
    recce_cloud.set_onboarding_state(new_state)



from .check import CheckDAO
from .run import RunDAO
from .types import Check, Run, RunProgress, RunType

# Explicitly declare exports
__all__ = ["CheckDAO", "RunDAO", "Check", "Run", "RunProgress", "RunType"]



"""
Artifact utilities for recce-cloud.

Simplified version of recce.artifact with only the functions needed
for upload-session functionality.
"""

import json
import os


def verify_artifacts_path(target_path: str) -> bool:
    """
    Verify if the target path contains valid dbt artifacts.

    Args:
        target_path: Path to the directory containing artifacts

    Returns:
        True if the target path contains manifest.json and catalog.json
    """
    if not target_path:
        return False

    if not os.path.exists(target_path):
        return False

    if not os.path.isdir(target_path):
        return False

    required_artifacts_files = ["manifest.json", "catalog.json"]

    if all(f in os.listdir(target_path) for f in required_artifacts_files):
        return True

    return False


def get_adapter_type(manifest_path: str) -> str:
    """
    Extract adapter type from manifest.json.

    Args:
        manifest_path: Path to manifest.json file

    Returns:
        Adapter type string (e.g., "postgres", "snowflake", "bigquery")

    Raises:
        Exception: If adapter type cannot be found in manifest
    """
    with open(manifest_path, "r", encoding="utf-8") as f:
        manifest_data = json.load(f)
        adapter_type = manifest_data.get("metadata", {}).get("adapter_type")
        if adapter_type is None:
            raise Exception("Failed to parse adapter type from manifest.json")
        return adapter_type



import difflib

import pandas as pd


def diff_text(before: str, after: str):
    if before is None and after is None:
        print("not found in both states")
        return
    elif before == after:
        print("no changes")
        return

    diff_output = difflib.unified_diff(before.splitlines(), after.splitlines(), "base", "current", lineterm="")
    for line in diff_output:
        print(line)


def diff_dataframe(before: pd.DataFrame, after: pd.DataFrame):
    if before is None and after is None:
        print("not found in both states")
        return

    before_aligned, after_aligned = before.align(after)
    diff = before_aligned.compare(after_aligned, result_names=("base", "current"))
    print(diff.to_string(na_rep="-"))



import os
from unittest import TestCase
from unittest.mock import MagicMock

import pytest

from recce.adapter.dbt_adapter import DbtAdapter, load_catalog, load_manifest
from recce.core import RecceContext, set_default_context
from recce.run import schema_diff_should_be_approved


def test_validator():
    from recce.tasks.schema import SchemaDiffCheckValidator

    validator = SchemaDiffCheckValidator()

    def validate(params: dict):
        validator.validate(
            {
                "name": "test",
                "type": "schema_diff",
                "params": params,
            }
        )

    # Select all models
    validate({})

    # Select by node name
    validate(
        {
            "node_id": "abc",
        }
    )
    validate(
        {
            "node_id": ["abc"],
        }
    )

    # Select by selector
    validate(
        {
            "select": "customers",
            "exclude": "customers",
            "packages": ["jaffle_shop"],
            "view_mode": "all",
        }
    )

    # packages should be an array
    with pytest.raises(ValueError):
        validate(
            {
                "packages": "jaffle_shop",
            }
        )

    # view_mode should be 'all' or 'changed_models'
    validate(
        {
            "view_mode": None,
        }
    )
    validate(
        {
            "view_mode": "all",
        }
    )
    with pytest.raises(ValueError):
        validate(
            {
                "view_mode": "abc",
            }
        )


test_root_path = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))


class TestSchemaDiffAutoApprove(TestCase):

    def setUp(self):
        self.default_context = MagicMock(spec=RecceContext)
        manifest = load_manifest(path=os.path.join(test_root_path, "manifest.json"))
        catalog = load_catalog(path=os.path.join(test_root_path, "catalog.json"))
        dbt_adapter = DbtAdapter(curr_manifest=manifest, curr_catalog=catalog)
        self.default_context.adapter = dbt_adapter

        dbt_adapter.adapter = MagicMock()
        dbt_adapter.adapter.type.return_value = None

        dbt_adapter.select_nodes = MagicMock()
        # Base and Current will be the same
        self.default_context.get_lineage.return_value = dbt_adapter.get_lineage()
        set_default_context(self.default_context)

    def test_schema_diff_should_be_approved(self):
        # Node_id is string
        is_approved = schema_diff_should_be_approved(
            {
                "node_id": "model.jaffle_shop.customers",
            }
        )
        assert is_approved is True

        # Node_id is list
        is_approved = schema_diff_should_be_approved(
            {
                "node_id": ["model.jaffle_shop.customers"],
            }
        )
        assert is_approved is True

        # Select all models
        self.default_context.adapter.select_nodes.return_value = ["model.jaffle_shop.customers"]
        is_approved = schema_diff_should_be_approved(
            {
                "select": "customers",
            }
        )
        assert is_approved is True



import os
import unittest

from recce.adapter.dbt_adapter import DbtAdapter, DbtVersion, load_manifest
from recce.core import RecceContext, set_default_context
from recce.summary import (
    _build_lineage_graph,
    generate_mermaid_lineage_graph,
    generate_summary_metadata,
)

current_dir = os.path.dirname(os.path.abspath(__file__))
base_manifest_dir = os.path.join(current_dir, "data", "manifest", "base")
pr2_manifest_dir = os.path.join(current_dir, "data", "manifest", "pr2")  # Pull Request 2l

dbt_version = DbtVersion()


@unittest.skipIf(dbt_version < "1.8.1", "Dbt version is less than 1.8.1")
def test_generate_summary_metadata():
    manifest = load_manifest(path=os.path.join(current_dir, "manifest.json"))
    assert manifest is not None
    dbt_adapter = DbtAdapter(curr_manifest=manifest)
    curr_lineage = dbt_adapter.get_lineage()
    base_lineage = dbt_adapter.get_lineage()

    # Summary with no changes
    generate_summary_metadata(curr_lineage, base_lineage)

    base_manifest = load_manifest(path=os.path.join(base_manifest_dir, "manifest.json"))
    curr_manifest = load_manifest(path=os.path.join(pr2_manifest_dir, "manifest.json"))
    dbt_adapter = DbtAdapter(curr_manifest=curr_manifest, base_manifest=base_manifest)
    curr_lineage = dbt_adapter.get_lineage()
    base_lineage = dbt_adapter.get_lineage(base=True)
    generate_summary_metadata(curr_lineage, base_lineage)


@unittest.skipIf(dbt_version < "v1.8.1", "Dbt version is less than 1.8.1")
def test_build_lineage_graph():
    base_manifest = load_manifest(path=os.path.join(base_manifest_dir, "manifest.json"))
    curr_manifest = load_manifest(path=os.path.join(pr2_manifest_dir, "manifest.json"))
    dbt_adapter = DbtAdapter(curr_manifest=curr_manifest, base_manifest=base_manifest)
    curr_lineage = dbt_adapter.get_lineage()
    base_lineage = dbt_adapter.get_lineage(base=True)

    lineage_graph = _build_lineage_graph(curr_lineage, base_lineage)
    # Modified nodes: 3
    # - int_customer_orders: added
    # - int_customer_payments: added
    # - customers: modified
    assert len(lineage_graph.modified_set) == 3


@unittest.skipIf(dbt_version < "v1.8.1", "Dbt version is less than 1.8.1")
def test_generate_mermaid_lineage_graph():
    set_default_context(RecceContext())
    base_manifest = load_manifest(path=os.path.join(base_manifest_dir, "manifest.json"))
    curr_manifest = load_manifest(path=os.path.join(pr2_manifest_dir, "manifest.json"))
    dbt_adapter = DbtAdapter(curr_manifest=curr_manifest, base_manifest=base_manifest)
    curr_lineage = dbt_adapter.get_lineage()
    base_lineage = dbt_adapter.get_lineage(base=True)
    graph = _build_lineage_graph(curr_lineage, base_lineage)
    mermaid_content, is_empty_graph, is_partial_graph = generate_mermaid_lineage_graph(graph)
    assert is_empty_graph is False
    assert is_partial_graph is False



from typing import List, Literal, Optional, Union

from pydantic import BaseModel

from recce.core import default_context
from recce.models import Check
from recce.tasks import Task
from recce.tasks.core import CheckValidator, TaskResultDiffer
from recce.tasks.query import QueryMixin


class RowCountParams(BaseModel):
    node_names: Optional[list[str]] = None
    node_ids: Optional[list[str]] = None


class RowCountTask(Task, QueryMixin):
    def __init__(self, params: dict):
        super().__init__()
        self.params = RowCountParams(**params) if params is not None else RowCountParams()
        self.connection = None

    def _query_row_count(self, dbt_adapter, model_name, base=False):
        node = dbt_adapter.find_node_by_name(model_name, base=base)
        if node is None:
            return None

        if node.resource_type != "model" and node.resource_type != "snapshot":
            return None

        if node.config and node.config.materialized not in ["table", "view", "incremental", "snapshot"]:
            return None

        relation = dbt_adapter.create_relation(model_name, base=base)
        if relation is None:
            return None

        sql_template = r"select count(*) from {{ relation }}"
        sql = dbt_adapter.generate_sql(sql_template, context=dict(relation=relation))
        _, table = dbt_adapter.execute(sql, fetch=True)
        return int(table[0][0]) if table[0][0] is not None else 0

    def execute(self):
        result = {}

        dbt_adapter = default_context().adapter

        query_candidates = []
        if self.params.node_ids or self.params.node_names:
            for node_id in self.params.node_ids or []:
                name = dbt_adapter.get_node_name_by_id(node_id)
                if name:
                    query_candidates.append(name)
            for node in self.params.node_names or []:
                query_candidates.append(node)
        else:

            def countable(unique_id):
                return unique_id.startswith("model") or unique_id.startswith("snapshot") or unique_id.startswith("seed")

            node_ids = dbt_adapter.select_nodes(
                select=self.params.select,
                exclude=self.params.exclude,
                packages=self.params.packages,
                view_mode=self.params.view_mode,
            )
            node_ids = list(filter(countable, node_ids))
            for node_id in node_ids:
                name = dbt_adapter.get_node_name_by_id(node_id)
                if name:
                    query_candidates.append(name)

        # Query row count for nodes that are not cached
        with dbt_adapter.connection_named("query"):
            self.connection = dbt_adapter.get_thread_connection()
            completed = 0
            total = len(query_candidates)
            for node in query_candidates:
                self.update_progress(message=f"Query: {node} [{completed}/{total}]", percentage=completed / total)

                row_count = self._query_row_count(dbt_adapter, node, base=False)
                self.check_cancel()
                result[node] = {
                    "curr": row_count,
                }
                completed += 1

        return result

    def cancel(self):
        super().cancel()
        if self.connection:
            self.close_connection(self.connection)


class RowCountDiffParams(BaseModel):
    node_names: Optional[list[str]] = None
    node_ids: Optional[list[str]] = None
    select: Optional[str] = None
    exclude: Optional[str] = None
    packages: Optional[list[str]] = None
    view_mode: Optional[Literal["all", "changed_models"]] = None


class RowCountDiffTask(Task, QueryMixin):
    def __init__(self, params: dict):
        super().__init__()
        self.params = RowCountDiffParams(**params) if params is not None else RowCountDiffParams()
        self.connection = None

    def _query_row_count(self, dbt_adapter, model_name, base=False):
        node = dbt_adapter.find_node_by_name(model_name, base=base)
        if node is None:
            return None

        if node.resource_type != "model" and node.resource_type != "snapshot":
            return None

        if node.config and node.config.materialized not in ["table", "view", "incremental", "snapshot"]:
            return None

        relation = dbt_adapter.create_relation(model_name, base=base)
        if relation is None:
            return None

        sql_template = r"select count(*) from {{ relation }}"
        sql = dbt_adapter.generate_sql(sql_template, context=dict(relation=relation))
        _, table = dbt_adapter.execute(sql, fetch=True)
        return int(table[0][0]) if table[0][0] is not None else 0

    def execute_dbt(self):
        result = {}

        dbt_adapter = default_context().adapter

        query_candidates = []
        if self.params.node_ids or self.params.node_names:
            for node_id in self.params.node_ids or []:
                name = dbt_adapter.get_node_name_by_id(node_id)
                if name:
                    query_candidates.append(name)
            for node in self.params.node_names or []:
                query_candidates.append(node)
        else:

            def countable(unique_id):
                return unique_id.startswith("model") or unique_id.startswith("snapshot") or unique_id.startswith("seed")

            node_ids = dbt_adapter.select_nodes(
                select=self.params.select,
                exclude=self.params.exclude,
                packages=self.params.packages,
                view_mode=self.params.view_mode,
            )
            node_ids = list(filter(countable, node_ids))
            for node_id in node_ids:
                name = dbt_adapter.get_node_name_by_id(node_id)
                if name:
                    query_candidates.append(name)

        # Query row count for nodes that are not cached
        with dbt_adapter.connection_named("query"):
            self.connection = dbt_adapter.get_thread_connection()
            completed = 0
            total = len(query_candidates)
            for node in query_candidates:
                self.update_progress(message=f"Diff: {node} [{completed}/{total}]", percentage=completed / total)

                base_row_count = self._query_row_count(dbt_adapter, node, base=True)
                self.check_cancel()
                curr_row_count = self._query_row_count(dbt_adapter, node, base=False)
                self.check_cancel()
                result[node] = {
                    "base": base_row_count,
                    "curr": curr_row_count,
                }
                completed += 1

        return result

    def execute_sqlmesh(self):
        result = {}

        query_candidates = []

        for node_id in self.node_ids or []:
            query_candidates.append(node_id)
        for node_name in self.params.node_names or []:
            query_candidates.append(node_name)

        from recce.adapter.sqlmesh_adapter import SqlmeshAdapter

        sqlmesh_adapter: SqlmeshAdapter = default_context().adapter

        for name in query_candidates:
            base_row_count = None
            curr_row_count = None

            try:
                df, _ = sqlmesh_adapter.fetchdf_with_limit(f"select count(*) from {name}", base=True)
                base_row_count = int(df.iloc[0, 0])
            except Exception:
                pass
            self.check_cancel()

            try:
                df, _ = sqlmesh_adapter.fetchdf_with_limit(f"select count(*) from {name}", base=False)
                curr_row_count = int(df.iloc[0, 0])
            except Exception:
                pass
            self.check_cancel()
            result[name] = {
                "base": base_row_count,
                "curr": curr_row_count,
            }

        return result

    def execute(self):
        context = default_context()
        if context.adapter_type == "dbt":
            return self.execute_dbt()
        else:
            return self.execute_sqlmesh()

    def cancel(self):
        super().cancel()
        if self.connection:
            self.close_connection(self.connection)


class RowCountDiffResultDiffer(TaskResultDiffer):
    def _check_result_changed_fn(self, result):
        base = {}
        current = {}

        for node, row_counts in result.items():
            base[node] = row_counts["base"]
            current[node] = row_counts["curr"]

        return TaskResultDiffer.diff(base, current)

    def _get_related_node_ids(self) -> Union[List[str], None]:
        """
        Get the related node ids.
        Should be implemented by subclass.
        """
        params = self.run.params
        if params.get("model"):
            return [TaskResultDiffer.get_node_id_by_name(params.get("model"))]
        elif params.get("node_names"):
            names = params.get("node_names", [])
            return [TaskResultDiffer.get_node_id_by_name(name) for name in names]
        elif params.get("node_ids"):
            return params.get("node_ids", [])
        else:
            return TaskResultDiffer.get_node_ids_by_selector(
                select=params.get("select"),
                exclude=params.get("exclude"),
                packages=params.get("packages"),
                view_mode=params.get("view_mode"),
            )

    def _get_changed_nodes(self) -> Union[List[str], None]:
        if self.changes:
            # Both affected_root_keys of deepdiff v7 (OrderedSet) and v8 (SetOrdered) are iterable
            # Convert to list directly
            return list(self.changes.affected_root_keys)
        return None


class RowCountDiffCheckValidator(CheckValidator):
    def validate_check(self, check: Check):
        try:
            RowCountDiffParams(**check.params)
        except Exception as e:
            raise ValueError(f"Invalid params: str{e}")



import pydantic
from pydantic import BaseModel


def pydantic_model_json_dump(model: BaseModel):
    pydantic_version = pydantic.version.VERSION
    pydantic_major = pydantic_version.split(".")[0]

    if pydantic_major == "1":
        return model.json(exclude_none=True)
    else:
        return model.model_dump_json(exclude_none=True)


def pydantic_model_dump(model: BaseModel):
    pydantic_version = pydantic.version.VERSION
    pydantic_major = pydantic_version.split(".")[0]

    if pydantic_major == "1":
        return model.dict()
    else:
        return model.model_dump()



#!/usr/bin/env python
"""
Recce Cloud CLI - Lightweight command for managing Recce Cloud operations.
"""

import os
import sys

import click
import requests
from rich.console import Console

from recce_cloud import __version__
from recce_cloud.api.client import RecceCloudClient, RecceCloudException
from recce_cloud.artifact import get_adapter_type, verify_artifacts_path


@click.group()
def cloud_cli():
    """
    Recce Cloud CLI - Manage Recce Cloud sessions and state files.

    A lightweight tool for CI/CD environments to interact with Recce Cloud
    without the heavy dependencies of the full recce package.
    """
    pass


@cloud_cli.command()
def version():
    """Show the version of recce-cloud."""
    click.echo(__version__)


@cloud_cli.command()
@click.option(
    "--target-path",
    type=click.Path(exists=True),
    default="target",
    help="Path to dbt target directory containing manifest.json and catalog.json",
)
@click.option(
    "--session-id",
    envvar="RECCE_SESSION_ID",
    help="Recce Cloud session ID to upload artifacts to (or use RECCE_SESSION_ID env var)",
)
def upload(target_path, session_id):
    """
    Upload dbt artifacts to Recce Cloud session.

    Lightweight replacement for 'recce upload-session' designed for CI/CD environments.
    Uploads manifest.json and catalog.json from the dbt target directory to a specific
    Recce Cloud session.

    \b
    What gets uploaded:
    - manifest.json: dbt project structure and model definitions
    - catalog.json: database catalog information and statistics

    \b
    The upload process:
    1. Validate dbt artifact files exist and are valid JSON
    2. Auto-detect Git repository and branch information
    3. Auto-detect Pull/Merge Request information (if in PR context)
    4. Auto-detect CI/CD platform information
    5. Authenticate with Recce Cloud API
    6. Upload artifacts to S3 for the session
    7. Update session metadata (links session to PR/MR via pr_link)

    \b
    About Recce Cloud Sessions:
    - Sessions compare base (production) and current (PR) environments
    - Each session stores manifests/catalogs from both environments
    - Sessions are linked to PRs/MRs for team collaboration and review
    - This enables automated PR gating and impact analysis in CI/CD

    \b
    Authentication:
    Uses environment variables for authentication:
    - RECCE_API_TOKEN: Recce Cloud API token (required)

    \b
    Auto-Detection:
    This command automatically detects your environment:
    - SCM Provider: GitHub, GitLab, Bitbucket, or generic git
    - CI/CD Platform: GitHub Actions, GitLab CI, Bitbucket Pipelines, CircleCI, etc.
    - Git Information: repository, branch, commit hash
    - PR/MR Information: ID, title, URL (used to link session to PR/MR)

    \b
    Environment Variables:
    - RECCE_SESSION_ID: Target session ID for upload (required)
    - RECCE_API_TOKEN: Recce Cloud API token (required)

    \b
    Examples:
    # Upload from default target directory with session ID from env
    export RECCE_SESSION_ID=abc123
    recce-cloud upload

    \b
    # Upload from custom target path
    recce-cloud upload --target-path my-target --session-id abc123

    \b
    # Typical CI/CD usage (GitHub Actions)
    recce-cloud upload --session-id ${{ steps.create-session.outputs.session-id }}

    \b
    Exit Codes:
    0 - Success
    1 - Environment detection error
    2 - Authentication error
    3 - File validation error (missing or invalid manifest/catalog)
    4 - Upload error
    """
    console = Console()

    # Validate session ID
    if not session_id:
        console.print("[red]Error:[/red] Session ID is required")
        console.print("Provide --session-id or set RECCE_SESSION_ID environment variable")
        sys.exit(2)

    # 1. Validate artifacts exist
    if not verify_artifacts_path(target_path):
        console.print(f"[red]Error:[/red] Invalid target path: {target_path}")
        console.print("Please provide a valid target path containing manifest.json and catalog.json.")
        sys.exit(3)

    manifest_path = os.path.join(target_path, "manifest.json")
    catalog_path = os.path.join(target_path, "catalog.json")

    # 2. Extract adapter type from manifest
    try:
        adapter_type = get_adapter_type(manifest_path)
    except Exception as e:
        console.print("[red]Error:[/red] Failed to parse adapter type from manifest.json")
        console.print(f"Reason: {e}")
        sys.exit(3)

    # 3. Get authentication token
    token = os.getenv("RECCE_API_TOKEN")
    if not token:
        console.print("[red]Error:[/red] No authentication token provided")
        console.print("Set RECCE_API_TOKEN environment variable")
        sys.exit(2)

    # 4. Initialize API client
    try:
        client = RecceCloudClient(token)
    except Exception as e:
        console.print("[red]Error:[/red] Failed to initialize API client")
        console.print(f"Reason: {e}")
        sys.exit(2)

    # 5. Get session info (org_id, project_id)
    console.print(f'Uploading artifacts for session ID "{session_id}"')
    try:
        session = client.get_session(session_id)
        if session.get("status") == "error":
            console.print(f"[red]Error:[/red] {session.get('message')}")
            sys.exit(2)

        org_id = session.get("org_id")
        if org_id is None:
            console.print(f"[red]Error:[/red] Session ID {session_id} does not belong to any organization.")
            sys.exit(2)

        project_id = session.get("project_id")
        if project_id is None:
            console.print(f"[red]Error:[/red] Session ID {session_id} does not belong to any project.")
            sys.exit(2)

    except RecceCloudException as e:
        console.print("[red]Error:[/red] Failed to get session info")
        console.print(f"Reason: {e.reason}")
        sys.exit(2)
    except Exception as e:
        console.print("[red]Error:[/red] Failed to get session info")
        console.print(f"Reason: {e}")
        sys.exit(2)

    # 6. Get presigned URLs
    try:
        presigned_urls = client.get_upload_urls_by_session_id(org_id, project_id, session_id)
    except RecceCloudException as e:
        console.print("[red]Error:[/red] Failed to get upload URLs")
        console.print(f"Reason: {e.reason}")
        sys.exit(4)
    except Exception as e:
        console.print("[red]Error:[/red] Failed to get upload URLs")
        console.print(f"Reason: {e}")
        sys.exit(4)

    # 7. Upload manifest.json
    console.print(f'Uploading manifest from path "{manifest_path}"')
    try:
        with open(manifest_path, "rb") as f:
            response = requests.put(presigned_urls["manifest_url"], data=f.read())
        if response.status_code not in [200, 204]:
            raise Exception(f"Upload failed with status {response.status_code}: {response.text}")
    except Exception as e:
        console.print("[red]Error:[/red] Failed to upload manifest.json")
        console.print(f"Reason: {e}")
        sys.exit(4)

    # 8. Upload catalog.json
    console.print(f'Uploading catalog from path "{catalog_path}"')
    try:
        with open(catalog_path, "rb") as f:
            response = requests.put(presigned_urls["catalog_url"], data=f.read())
        if response.status_code not in [200, 204]:
            raise Exception(f"Upload failed with status {response.status_code}: {response.text}")
    except Exception as e:
        console.print("[red]Error:[/red] Failed to upload catalog.json")
        console.print(f"Reason: {e}")
        sys.exit(4)

    # 9. Update session metadata
    try:
        client.update_session(org_id, project_id, session_id, adapter_type)
    except RecceCloudException as e:
        console.print("[red]Error:[/red] Failed to update session metadata")
        console.print(f"Reason: {e.reason}")
        sys.exit(4)
    except Exception as e:
        console.print("[red]Error:[/red] Failed to update session metadata")
        console.print(f"Reason: {e}")
        sys.exit(4)

    # Success!
    console.rule("Uploaded Successfully", style="green")
    console.print(
        f'Uploaded dbt artifacts to Recce Cloud for session ID "{session_id}" from "{os.path.abspath(target_path)}"'
    )
    sys.exit(0)


if __name__ == "__main__":
    cloud_cli()



import pytest
from jinja2 import Template
from sqlglot import parse_one

from recce.tasks import ProfileDiffTask, ProfileTask
from recce.tasks.profile import PROFILE_COLUMN_JINJA_TEMPLATE

csv_data_curr = """
        customer_id,name,age
        1,Alice,30
        2,Bob,25
        3,Charlie,35
        """

csv_data_base = """
    customer_id,name,age
    1,Alice,35
    2,Bob,25
    3,Charlie,35
    """


def test_profile(dbt_test_helper):
    dbt_test_helper.create_model("customers", None, csv_data_curr)
    params = dict(model="customers")
    task = ProfileTask(params)
    run_result = task.execute()

    assert len(run_result.current.data) == 3


def test_profile_with_selected_columns(dbt_test_helper):
    dbt_test_helper.create_model("customers", None, csv_data_curr)
    params = dict(model="customers", columns=["name", "age"])
    task = ProfileTask(params)
    run_result = task.execute()
    assert len(run_result.current.data) == 2


def test_profile_diff(dbt_test_helper):
    dbt_test_helper.create_model("customers", csv_data_base, csv_data_curr)
    params = dict(model="customers")
    task = ProfileDiffTask(params)
    run_result = task.execute()

    assert len(run_result.current.data) == 3
    assert len(run_result.base.data) == 3


def test_profile_diff_with_selected_columns(dbt_test_helper):
    dbt_test_helper.create_model("customers", csv_data_base, csv_data_curr)
    params = dict(model="customers", columns=["name", "age"])
    task = ProfileDiffTask(params)
    run_result = task.execute()
    assert len(run_result.current.data) == 2
    assert len(run_result.base.data) == 2


def test_validator():
    from recce.tasks.profile import ProfileCheckValidator

    def validate(params: dict = {}, view_options: dict = {}):
        ProfileCheckValidator().validate(
            {
                "name": "test",
                "type": "profile_diff",
                "params": params,
                "view_options": view_options,
            }
        )

    validate(
        {
            "model": "customers",
        }
    )

    with pytest.raises(ValueError):
        validate({})


def test_profile_column_jinja_template():

    class DummyAdapter:
        def __init__(self, database_type="duckdb"):
            self.database_type = database_type

        def quote(self, col):
            quote_marks = {
                "athena": '"',
                "bigquery": "`",
                "databricks": "`",
                "duckdb": '"',
                "postgres": '"',
                "redshift": '"',
                "snowflake": '"',
                "sqlite": '"',
                "sqlserver": '"',
                "trino": '"',
            }
            quote_mark = quote_marks.get(self.database_type, '"')
            return f"{quote_mark}{col}{quote_mark}"

    class DummyDbt:
        def __init__(self, database_type="Postgres"):
            self.database_type = database_type

        def type_bigint(self):
            bigint_types = {
                "athena": "BIGINT",
                "bigquery": "INT64",
                "databricks": "BIGINT",
                "duckdb": "BIGINT",
                "postgres": "BIGINT",
                "redshift": "BIGINT",
                "snowflake": "NUMBER",
                "sqlite": "INTEGER",
                "sqlserver": "BIGINT",
                "trino": "BIGINT",
            }
            return bigint_types.get(self.database_type, "bigint")

        def type_numeric(self):
            numeric_types = {
                "athena": "DECIMAL",
                "bigquery": "NUMERIC",
                "databricks": "DECIMAL",
                "duckdb": "DECIMAL",
                "postgres": "NUMERIC",
                "redshift": "DECIMAL",
                "snowflake": "NUMBER",
                "sqlite": "NUMERIC",
                "sqlserver": "DECIMAL",
                "trino": "DECIMAL",
            }
            return numeric_types.get(self.database_type, "numeric")

        def type_string(self):
            string_types = {
                "athena": "VARCHAR",
                "bigquery": "STRING",
                "databricks": "STRING",
                "duckdb": "VARCHAR",
                "postgres": "TEXT",
                "redshift": "VARCHAR",
                "snowflake": "VARCHAR",
                "sqlite": "TEXT",
                "sqlserver": "VARCHAR",
                "trino": "VARCHAR",
            }
            return string_types.get(self.database_type, "varchar")

    test_db_column_types = {
        "athena": [
            "boolean",
            "int",
            "double",
            "decimal",
            "char",
            "string",
            "binary",
            "date",
            "array",
            "struct",
            "map",
        ],
        "bigquery": [
            "array",
            "bool",
            "bytes",
            "date",
            "datetime",
            "interval",
            "json",
            "int64",
            "bignumeric",
            "range",
            "string",
            "struct",
            "time",
            "timestamp",
        ],
        "databricks": [
            "bigint",
            "binary",
            "boolean",
            "date",
            "decimal",
            "double",
            "float",
            "int",
            "interval",
            "void",
            "string",
            "timestamp",
            "array",
            "map",
            "struct",
            "variant",
            "object",
        ],
        "duckdb": [
            "boolean",
            "integer",
            "bigint",
            "real",
            "double",
            "decimal",
            "varchar",
            "date",
            "time",
            "timestamp",
            "blob",
            "array",
            "list",
            "struct",
            "map",
            "union",
            "uuid",
            "json",
        ],
        "postgres": [
            "bigint",
            "bigserial",
            "bit",
            "boolean",
            "box",
            "bytea",
            "character",
            "cidr",
            "circle",
            "date",
            "double precision",
            "inet",
            "integer",
            "interval",
            "json",
            "jsonb",
            "line",
            "lseg",
            "macaddr",
            "money",
            "numeric",
            "path",
            "pg_lsn",
            "point",
            "polygon",
            "real",
            "smallint",
            "smallserial",
            "serial",
            "text",
            "time",
            "timestamp",
            "tsquery",
            "tsvector",
            "txid_snapshot",
            "uuid",
            "xml",
            "array",
        ],
        "redshift": [
            "smallint",
            "integer",
            "bigint",
            "decimal",
            "real",
            "double precision",
            "boolean",
            "char",
            "varchar",
            "date",
            "timestamp",
            "timestamptz",
            "super",
            "time",
            "timetz",
            "varbyte",
            "geometry",
            "geography",
            "hllsketch",
        ],
        "snowflake": [
            "number",
            "decimal",
            "numeric",
            "int",
            "integer",
            "bigint",
            "smallint",
            "tinyint",
            "byteint",
            "float",
            "float4",
            "float8",
            "double",
            "double precision",
            "real",
            "varchar",
            "char",
            "character",
            "string",
            "text",
            "binary",
            "varbinary",
            "boolean",
            "date",
            "datetime",
            "time",
            "timestamp",
            "timestamp_ltz",
            "timestamp_ntz",
            "timestamp_tz",
            "variant",
            "object",
            "array",
            "geography",
            "geometry",
        ],
        "sqlite": ["integer", "real", "text", "blob", "numeric", "boolean", "date", "datetime", "json"],
        "trino": [
            "boolean",
            "tinyint",
            "smallint",
            "integer",
            "bigint",
            "real",
            "double",
            "decimal",
            "varchar",
            "char",
            "varbinary",
            "json",
            "date",
            "time",
            "timestamp",
            "timestamp with time zone",
            "interval year to month",
            "interval day to second",
            "array",
            "map",
            "row",
            "ipaddress",
            "uuid",
            "hyperloglog",
            "p4hyperloglog",
            "qdigest",
        ],
        "sqlserver": [
            "bigint",
            "int",
            "smallint",
            "tinyint",
            "bit",
            "decimal",
            "numeric",
            "money",
            "smallmoney",
            "float",
            "real",
            "date",
            "datetime",
            "datetime2",
            "smalldatetime",
            "time",
            "char",
            "varchar",
            "text",
            "nchar",
            "nvarchar",
            "ntext",
            "binary",
            "varbinary",
            "image",
            "timestamp",
            "uniqueidentifier",
            "xml",
            "cursor",
            "table",
            "sql_variant",
        ],
    }

    for db_type, column_types in test_db_column_types.items():
        for column_type in column_types:
            context = {
                "column_type": column_type,
                "column_name": "profile_column",
                "db_type": db_type,
                "relation": "test_table",
                "adapter": DummyAdapter(),
                "dbt": DummyDbt(),
            }

            sql = Template(PROFILE_COLUMN_JINJA_TEMPLATE).render(context)
            dialect = db_type if db_type != "sqlserver" else "tsql"
            parse_one(sql, read=dialect)



from typing import List

from pydantic import BaseModel

from ..core import default_context
from ..exceptions import RecceException
from ..models import Check
from .core import CheckValidator, Task, TaskResultDiffer
from .dataframe import DataFrame

PROFILE_COLUMN_JINJA_TEMPLATE = r"""
{# Conditions -------------------------------------------- #}
{%- set is_struct = column_type.startswith('struct') -%}
{%- set is_numeric =
    column_type.startswith('int') or
    column_type.startswith('float') or
    'numeric' in column_type or
    'number' in column_type or
    'double' in column_type or
    'bigint' in column_type
-%}
{%- set is_date_or_time =
    column_type.startswith('date') or
    column_type.startswith('timestamp')
-%}
{%- set is_logical = column_type.startswith('bool') -%}

{%- if db_type == 'sqlserver' -%}
    {%- set is_numeric = column_type in [
        "bigint", "numeric", "smallint", "decimal", "int",
        "tinyint", "money", "float", "real"
    ]-%}
{%- elif db_type == 'athena' -%}
    {%- set is_numeric =
        "int" in column_type or
        "float" in column_type or
        "decimal" in column_type or
        "double" in column_type
    -%}
{%- endif -%}

{# General Agg ------------------------------------------- #}
{%- set agg_row_count = 'cast(count(*) as ' ~ dbt.type_bigint() ~ ')' -%}
{%- set agg_not_null_proportion =
        'sum(case when ' ~ adapter.quote(column_name) ~ ' is null '
        ~ 'then 0 '
        ~ 'else 1 end) / '
        ~ 'cast(count(*) as ' ~ dbt.type_numeric() ~ ')'
-%}
{%- set agg_distinct_proportion =
        'count(distinct ' ~ adapter.quote(column_name) ~') / '
        ~ 'cast(count(*) as ' ~ dbt.type_numeric() ~ ')'
-%}
{%- set agg_distinct_count = 'count(distinct ' ~ adapter.quote(column_name) ~ ')' -%}
{%- set agg_is_unique =      'count(distinct ' ~ adapter.quote(column_name) ~ ') = count(*)' -%}
{%- set agg_min =            'cast(null as ' ~ dbt.type_string() ~ ')' -%}
{%- set agg_max =            'cast(null as ' ~ dbt.type_string() ~ ')' -%}
{%- set agg_avg =            'cast(null as ' ~ dbt.type_numeric() ~ ')' -%}
{%- set agg_median =         'cast(null as ' ~ dbt.type_numeric() ~ ')' -%}


{%- if is_struct -%}
    {%- set agg_distinct_proportion = 'cast(null as ' ~ dbt.type_numeric() ~ ')' -%}
    {%- set agg_distinct_count = 'cast(null as ' ~ dbt.type_numeric() ~ ')' -%}
    {%- set agg_is_unique = 'null' -%}
{%- endif -%}


{%- if (is_numeric or is_date_or_time) and (not is_struct) -%}
    {%- set agg_min =
        'cast(min(' ~ adapter.quote(column_name) ~ ') as ' ~ dbt.type_string() ~ ')'
    -%}
    {%- set agg_max =
        'cast(max(' ~ adapter.quote(column_name) ~ ') as ' ~ dbt.type_string() ~ ')'
    -%}
{%- endif -%}


{%- if is_numeric and not is_struct -%}
    {%- set agg_avg = 'avg(' ~ adapter.quote(column_name) ~ ')' -%}

    {%- if db_type == 'bigquery' -%}
        {%- set agg_median = 'approx_quantiles(' ~ adapter.quote(column_name) ~ ', 100)[offset(50)]' -%}
    {%- elif db_type == 'postgres' -%}
        {%- set agg_median = 'percentile_cont(0.5) within group (order by ' ~ adapter.quote(column_name) ~ ')' -%}
    {%- elif db_type == 'redshift' -%}
        {%- set agg_median =
            '(select percentile_cont(0.5) within group (order by '
            ~ adapter.quote(column_name) ~ ') from ' ~ relation ~ ')' -%}
    {%- elif db_type == 'athena' -%}
        {%- set agg_median = 'approx_percentile( ' ~ adapter.quote(column_name) ~ ', 0.5)' -%}
    {%- elif db_type == 'sqlserver' -%}
        {%- set agg_median = 'percentile_cont(' ~ adapter.quote(column_name) ~ ', 0.5) over ()' -%}
    {%- else -%}
        {%- set agg_median = 'median(' ~ adapter.quote(column_name) ~ ')' -%}
    {%- endif -%}
{%- elif is_logical -%}
    {%- set agg_avg = 'avg(case when ' ~ adapter.quote(column_name) ~ ' then 1 else 0 end)' -%}
{%- endif -%}


{# Overwrite Agg ----------------------------------------- #}

{# DRC-663: Support bigquery array type }
{%- set is_array = column_type.startswith('array') -%}
{%- if db_type == 'bigquery' and is_array -%}
    {%- set agg_distinct_proportion = 'cast(null as ' ~ dbt.type_numeric() ~ ')' -%}
    {%- set agg_distinct_count = 'cast(null as ' ~ dbt.type_numeric() ~ ')' -%}
    {%- set agg_is_unique = 'null' -%}
    {%- set agg_min =
        'cast(min(array_length(' ~ adapter.quote(column_name) ~ ')) as ' ~ dbt.type_string() ~ ')'
    -%}
    {%- set agg_max =
        'cast(max(array_length(' ~ adapter.quote(column_name) ~ ')) as ' ~ dbt.type_string() ~ ')'
    -%}
    {%- set agg_avg = 'avg(array_length(' ~ adapter.quote(column_name) ~ '))' -%}
    {%- set agg_median =
        'approx_quantiles(array_length(' ~ adapter.quote(column_name) ~ '), 100)[offset(50)]'
    -%}
{%- endif -%}


{# Main Query -------------------------------------------- #}

select
    '{{ column_name }}' as column_name,
    nullif('{{ column_type }}', '') as data_type,
    {{ agg_row_count }} as row_count,
    {{ agg_not_null_proportion }} as not_null_proportion,
    {{ agg_distinct_proportion }} as distinct_proportion,
    {{ agg_distinct_count }} as distinct_count,
    {{ agg_is_unique }} as is_unique,
    {{ agg_min }} as min,
    {{ agg_max }} as max,
    {{ agg_avg }} as avg,
    {{ agg_median }} as median
from {{ relation }}
"""


class ProfileParams(BaseModel):
    model: str
    columns: List[str] = None


class ProfileDiffResult(BaseModel):
    base: DataFrame
    current: DataFrame


class ProfileResult(BaseModel):
    current: DataFrame


class ProfileDiffTask(Task):

    def __init__(self, params):
        super().__init__()
        self.params = ProfileParams(**params)
        self.connection = None

    def execute(self):
        import agate

        from recce.adapter.dbt_adapter import DbtAdapter, merge_tables

        dbt_adapter: DbtAdapter = default_context().adapter

        model: str = self.params.model
        selected_columns: List[str] = self.params.columns

        with dbt_adapter.connection_named("query"):
            self.connection = dbt_adapter.get_thread_connection()

            base_columns = [column for column in dbt_adapter.get_columns(model, base=True)]
            curr_columns = [column for column in dbt_adapter.get_columns(model, base=False)]

            if selected_columns:
                # Only profile the columns in the filter_columns
                base_columns = [column for column in base_columns if column.name in selected_columns]
                curr_columns = [column for column in curr_columns if column.name in selected_columns]

            total = len(base_columns) + len(curr_columns)
            completed = 0

            tables: List[agate.Table] = []

            for column in base_columns:
                self.update_progress(message=f"[Base] Profile column: {column.name}", percentage=completed / total)
                relation = dbt_adapter.create_relation(model, base=True)
                response, table = self._profile_column(dbt_adapter, relation, column)
                tables.append(table)
                completed = completed + 1
                self.check_cancel()
            base = DataFrame.from_agate(merge_tables(tables))

            tables: List[agate.Table] = []
            for column in curr_columns:
                self.update_progress(message=f"[Current] Profile column: {column.column}", percentage=completed / total)
                relation = dbt_adapter.create_relation(model, base=False)
                response, table = self._profile_column(dbt_adapter, relation, column)
                tables.append(table)
                completed = completed + 1
                self.check_cancel()
            current = DataFrame.from_agate(merge_tables(tables))

            if len(base.columns) == 0 and len(current.columns) != 0:
                base.columns = current.columns
            elif len(base.columns) != 0 and len(current.columns) == 0:
                current.columns = base.columns

            return ProfileDiffResult(base=base, current=current)

    def _profile_column(self, dbt_adapter, relation, column):
        column_name = column.name
        column_type = column.data_type.lower()
        db_type = dbt_adapter.adapter.type().lower()

        try:
            sql = dbt_adapter.generate_sql(
                PROFILE_COLUMN_JINJA_TEMPLATE,
                base=False,  # always false because we use the macro in current manifest
                context=dict(relation=relation, column_name=column_name, column_type=column_type, db_type=db_type),
            )
        except Exception as e:
            raise RecceException(f"Failed to generate SQL for profiling column: {column_name}") from e

        try:
            return dbt_adapter.execute(sql, fetch=True)
        except Exception as e:
            from recce.adapter.dbt_adapter import dbt_version

            if dbt_version < "v1.8":
                from dbt.exceptions import DbtDatabaseError
            else:
                from dbt_common.exceptions import DbtDatabaseError
            if isinstance(e, DbtDatabaseError):
                if str(e).find("100051") >= 0:
                    # Snowflake error '100051 (22012): Division by zero"'
                    e = RecceException("No profile diff result due to the model is empty.", False)
            raise e

    def cancel(self):
        super().cancel()

        if self.connection:
            from recce.adapter.dbt_adapter import DbtAdapter

            dbt_adapter: DbtAdapter = default_context().adapter
            with dbt_adapter.connection_named("cancel"):
                dbt_adapter.cancel(self.connection)


class ProfileDiffResultDiffer(TaskResultDiffer):
    def _check_result_changed_fn(self, result):
        return self.diff(result["base"], result["current"])


class ProfileCheckValidator(CheckValidator):

    def validate_check(self, check: Check):
        try:
            ProfileParams(**check.params)
        except Exception as e:
            raise ValueError(f"Invalid check: {str(e)}")


class ProfileTask(ProfileDiffTask):
    def execute(self):
        import agate

        from recce.adapter.dbt_adapter import DbtAdapter, merge_tables

        dbt_adapter: DbtAdapter = default_context().adapter

        model: str = self.params.model
        selected_columns: List[str] = self.params.columns

        with dbt_adapter.connection_named("query"):
            self.connection = dbt_adapter.get_thread_connection()
            curr_columns = [column for column in dbt_adapter.get_columns(model, base=False)]

            if selected_columns:
                curr_columns = [column for column in curr_columns if column.name in selected_columns]

            total = len(curr_columns)
            completed = 0

            tables: List[agate.Table] = []
            for column in curr_columns:
                self.update_progress(message=f"[Current] Profile column: {column.column}", percentage=completed / total)
                relation = dbt_adapter.create_relation(model, base=False)
                response, table = self._profile_column(dbt_adapter, relation, column)
                tables.append(table)
                completed = completed + 1
                self.check_cancel()
            current = DataFrame.from_agate(merge_tables(tables))
            return ProfileResult(current=current)



import os.path
from unittest import TestCase
from unittest.mock import patch

from recce.config import RecceConfig
from recce.util import SingletonMeta

test_root_path = os.path.dirname(os.path.abspath(__file__))


class RecceConfigTestCase(TestCase):
    def setUp(self):
        self.recce_config_path = os.path.join(test_root_path, "data", "config", "recce.yml")
        pass

    def tearDown(self):
        # Reset the SingletonMeta instances due to RecceConfig is a singleton
        SingletonMeta._instances = {}

    def test_load_recce_config(self):
        config = RecceConfig(self.recce_config_path)

        # Test data contains 2 checks
        preset_checks = config.config.get("checks")
        self.assertIsNotNone(preset_checks)
        self.assertIsInstance(preset_checks, list)
        self.assertEqual(len(preset_checks), 2)

    @patch("recce.config.RecceConfig.save")
    def test_recce_config_not_found(self, mock_save):
        default_config = RecceConfig("NOT_EXISTING_FILE")
        assert mock_save.called is True
        # Default config should be generated
        preset_checks = default_config.config.get("checks")
        self.assertIsNotNone(default_config.config)
        self.assertIsInstance(preset_checks, list)
        self.assertEqual(len(preset_checks), 2)

    @patch("recce.yaml.safe_load")
    def test_recce_config_null_checks(self, mock_yaml_safe_load):
        # mock to load a yaml file with null checks
        mock_yaml_safe_load.return_value = {"checks": None}
        RecceConfig(self.recce_config_path)



"""
Recce Cloud API client for lightweight operations.

Simplified version of recce.util.recce_cloud.RecceCloud with only
the methods needed for upload-session functionality.
"""

import json
import os

import requests

RECCE_CLOUD_API_HOST = os.environ.get("RECCE_CLOUD_API_HOST", "https://cloud.datarecce.io")

DOCKER_INTERNAL_URL_PREFIX = "http://host.docker.internal"
LOCALHOST_URL_PREFIX = "http://localhost"


class RecceCloudException(Exception):
    """Exception raised when Recce Cloud API returns an error."""

    def __init__(self, message: str, reason: str, status_code: int):
        super().__init__(message)
        self.status_code = status_code

        try:
            reason = json.loads(reason).get("detail", "")
        except json.JSONDecodeError:
            pass
        self.reason = reason


class RecceCloudClient:
    """
    Lightweight Recce Cloud API client.

    Supports authentication with Recce Cloud API token (starts with "rct-").
    """

    def __init__(self, token: str):
        if token is None:
            raise ValueError("Token cannot be None.")
        self.token = token
        self.base_url_v2 = f"{RECCE_CLOUD_API_HOST}/api/v2"

    def _request(self, method: str, url: str, headers: dict = None, **kwargs):
        """Make authenticated HTTP request to Recce Cloud API."""
        headers = {
            **(headers or {}),
            "Authorization": f"Bearer {self.token}",
        }
        return requests.request(method, url, headers=headers, **kwargs)

    def _replace_localhost_with_docker_internal(self, url: str) -> str:
        """Convert localhost URLs to docker internal URLs if running in Docker."""
        if url is None:
            return None
        if (
            os.environ.get("RECCE_SHARE_INSTANCE_ENV") == "docker"
            or os.environ.get("RECCE_TASK_INSTANCE_ENV") == "docker"
            or os.environ.get("RECCE_INSTANCE_ENV") == "docker"
        ):
            # For local development, convert the presigned URL from localhost to host.docker.internal
            if url.startswith(LOCALHOST_URL_PREFIX):
                return url.replace(LOCALHOST_URL_PREFIX, DOCKER_INTERNAL_URL_PREFIX)
        return url

    def get_session(self, session_id: str) -> dict:
        """
        Get session information from Recce Cloud.

        Args:
            session_id: The session ID to retrieve

        Returns:
            dict containing session information with keys:
                - org_id: Organization ID
                - project_id: Project ID
                - ... other session fields

        Raises:
            RecceCloudException: If the request fails
        """
        api_url = f"{self.base_url_v2}/sessions/{session_id}"
        response = self._request("GET", api_url)
        if response.status_code == 403:
            return {"status": "error", "message": response.json().get("detail")}
        if response.status_code != 200:
            raise RecceCloudException(
                message="Failed to get session from Recce Cloud.",
                reason=response.text,
                status_code=response.status_code,
            )
        data = response.json()
        if data["success"] is not True:
            raise RecceCloudException(
                message="Failed to get session from Recce Cloud.",
                reason=data.get("message", "Unknown error"),
                status_code=response.status_code,
            )
        return data["session"]

    def get_upload_urls_by_session_id(self, org_id: str, project_id: str, session_id: str) -> dict:
        """
        Get presigned S3 upload URLs for a session.

        Args:
            org_id: Organization ID
            project_id: Project ID
            session_id: Session ID

        Returns:
            dict with keys:
                - manifest_url: Presigned URL for uploading manifest.json
                - catalog_url: Presigned URL for uploading catalog.json

        Raises:
            RecceCloudException: If the request fails
        """
        api_url = f"{self.base_url_v2}/organizations/{org_id}/projects/{project_id}/sessions/{session_id}/upload-url"
        response = self._request("GET", api_url)
        if response.status_code != 200:
            raise RecceCloudException(
                message="Failed to get upload URLs for session from Recce Cloud.",
                reason=response.text,
                status_code=response.status_code,
            )
        data = response.json()
        if data["presigned_urls"] is None:
            raise RecceCloudException(
                message="No presigned URLs returned from the server.",
                reason="",
                status_code=404,
            )

        presigned_urls = data["presigned_urls"]
        for key, url in presigned_urls.items():
            presigned_urls[key] = self._replace_localhost_with_docker_internal(url)
        return presigned_urls

    def update_session(self, org_id: str, project_id: str, session_id: str, adapter_type: str) -> dict:
        """
        Update session metadata with adapter type.

        Args:
            org_id: Organization ID
            project_id: Project ID
            session_id: Session ID
            adapter_type: dbt adapter type (e.g., "postgres", "snowflake", "bigquery")

        Returns:
            dict containing updated session information

        Raises:
            RecceCloudException: If the request fails
        """
        api_url = f"{self.base_url_v2}/organizations/{org_id}/projects/{project_id}/sessions/{session_id}"
        data = {"adapter_type": adapter_type}
        response = self._request("PATCH", api_url, json=data)
        if response.status_code == 403:
            return {"status": "error", "message": response.json().get("detail")}
        if response.status_code != 200:
            raise RecceCloudException(
                message="Failed to update session in Recce Cloud.",
                reason=response.text,
                status_code=response.status_code,
            )
        return response.json()



from recce.adapter.dbt_adapter import DbtAdapter


def test_select(dbt_test_helper):
    csv_data_curr = """
        customer_id,name,age
        1,Alice,30
        2,Bob,25
        3,Charlie,35
        """

    csv_data_base = """
        customer_id,name,age
        1,Alice,35
        2,Bob,25
        3,Charlie,35
        """

    dbt_test_helper.create_model("customers_1", csv_data_base, csv_data_curr)
    dbt_test_helper.create_model("customers_2", csv_data_base, csv_data_base, depends_on=["customers_1"])
    adapter: DbtAdapter = dbt_test_helper.context.adapter

    # Test methods
    node_ids = adapter.select_nodes("customers_1")
    assert len(node_ids) == 1
    node_ids = adapter.select_nodes("resource_type:model")
    assert len(node_ids) == 2
    node_ids = adapter.select_nodes("tag:test_tag")
    assert len(node_ids) == 2
    node_ids = adapter.select_nodes("tag:test_tag2")
    assert len(node_ids) == 0
    node_ids = adapter.select_nodes("config.materialized:incremental")
    assert len(node_ids) == 0
    node_ids = adapter.select_nodes("config.materialized:table")
    assert len(node_ids) == 2

    # Test state method
    node_ids = adapter.select_nodes("state:modified")
    assert len(node_ids) == 1

    # Test set operation
    node_ids = adapter.select_nodes("customers_1 customers_2")
    assert len(node_ids) == 2
    node_ids = adapter.select_nodes("customers_1,customers_2")
    assert len(node_ids) == 0
    node_ids = adapter.select_nodes("config.materialized:table,tag:test_tag")
    assert len(node_ids) == 2
    node_ids = adapter.select_nodes("config.materialized:table,tag:test_tag2")
    assert len(node_ids) == 0
    node_ids = adapter.select_nodes(exclude="customers_1")
    assert len(node_ids) == 1
    node_ids = adapter.select_nodes("customers_1", exclude="customers_2")
    assert len(node_ids) == 1

    # Test graph operation
    node_ids = adapter.select_nodes("state:modified+")
    assert len(node_ids) == 2
    node_ids = adapter.select_nodes("+state:modified")
    assert len(node_ids) == 1

    # Test resource type: snapshot
    dbt_test_helper.create_snapshot("snapshot_1", csv_data_base, csv_data_curr)
    dbt_test_helper.create_model("use_snapshot", csv_data_base, csv_data_base, depends_on=["snapshot_1"])

    node_ids = adapter.select_nodes("resource_type:snapshot")
    assert len(node_ids) == 1

    node_ids = adapter.select_nodes("resource_type:snapshot+")
    assert len(node_ids) == 2

    node_ids = adapter.select_nodes("state:modified,resource_type:snapshot")
    assert len(node_ids) == 1


def test_select_removed_by_graph(dbt_test_helper):
    csv_data_curr = """
        customer_id,name,age
        1,Alice,30
        2,Bob,25
        3,Charlie,35
        """

    csv_data_base = """
        customer_id,name,age
        1,Alice,35
        2,Bob,25
        3,Charlie,35
        """

    dbt_test_helper.create_model("customers_1", csv_data_base, csv_data_curr)
    dbt_test_helper.create_model("customers_2", base_csv=csv_data_base, depends_on=["customers_1"])
    dbt_test_helper.create_model("customers_3", curr_csv=csv_data_curr, depends_on=["customers_1"])
    adapter: DbtAdapter = dbt_test_helper.context.adapter

    # Test graph operation
    node_ids = adapter.select_nodes("customers_1+")
    assert len(node_ids) == 3
    node_ids = adapter.select_nodes("+customers_2")
    assert len(node_ids) == 2
    node_ids = adapter.select_nodes("+customers_3")
    assert len(node_ids) == 2


def test_select_source_removed_by_graph(dbt_test_helper):
    csv_data_curr = """
        customer_id,name,age
        1,Alice,30
        2,Bob,25
        3,Charlie,35
        """

    csv_data_base = """
        customer_id,name,age
        1,Alice,35
        2,Bob,25
        3,Charlie,35
        """

    dbt_test_helper.create_model("customers_2", base_csv=csv_data_base)
    dbt_test_helper.create_model("customers_3", curr_csv=csv_data_curr)
    adapter: DbtAdapter = dbt_test_helper.context.adapter

    # Test graph operation
    node_ids = adapter.select_nodes("state:modified")
    assert len(node_ids) == 2


def test_select_with_disabled(dbt_test_helper):
    csv_data_curr = """
        customer_id,name,age
        1,Alice,30
        2,Bob,25
        3,Charlie,35
        """

    csv_data_base = """
        customer_id,name,age
        1,Alice,35
        2,Bob,25
        3,Charlie,35
        """

    dbt_test_helper.create_model("customers_1", csv_data_base, csv_data_curr)
    dbt_test_helper.create_model("customers_2", csv_data_base, csv_data_curr, disabled=True)
    adapter: DbtAdapter = dbt_test_helper.context.adapter

    # Test graph operation
    node_ids = adapter.select_nodes("customers_1+")
    assert len(node_ids) == 1


def test_select_with_pacakage_mode_include_exclude(dbt_test_helper):
    csv_data_curr = """
            customer_id,name,age
            1,Alice,30
            2,Bob,25
            3,Charlie,35
            """

    csv_data_base = """
            customer_id,name,age
            1,Alice,35
            2,Bob,25
            3,Charlie,35
            """

    """
    The diagram of the models:
    customers_1 ── customers_2 ── customers_3(*) ├── other_package.customers_4
                                                 └── customers_5

    Only customers_3 is changed
    """
    dbt_test_helper.create_model("customers_1", csv_data_base, csv_data_base)
    dbt_test_helper.create_model("customers_2", csv_data_base, csv_data_base, depends_on=["customers_1"])
    dbt_test_helper.create_model("customers_3", csv_data_base, csv_data_curr, depends_on=["customers_2"])
    dbt_test_helper.create_model(
        "customers_4", csv_data_base, csv_data_base, depends_on=["customers_3"], package_name="other_package"
    )
    dbt_test_helper.create_model("customers_5", csv_data_base, csv_data_base, depends_on=["customers_3"])

    adapter: DbtAdapter = dbt_test_helper.context.adapter

    node_ids = adapter.select_nodes(packages=["other_package"])
    assert len(node_ids) == 1

    node_ids = adapter.select_nodes(view_mode="changed_models")
    assert len(node_ids) == 4

    node_ids = adapter.select_nodes(view_mode="changed_models", packages=["other_package"])
    assert len(node_ids) == 1

    node_ids = adapter.select_nodes(view_mode="changed_models", packages=["other_package"], exclude="customers_4")
    assert len(node_ids) == 0
    node_ids = adapter.select_nodes(view_mode="changed_models", packages=["other_package"], select="customers_1+")
    assert len(node_ids) == 1
    node_ids = adapter.select_nodes(view_mode="changed_models", select="+customers_5")
    assert len(node_ids) == 3
    node_ids = adapter.select_nodes(view_mode="all", select="+customers_5")
    assert len(node_ids) == 4
    node_ids = adapter.select_nodes(select="+customers_5")
    assert len(node_ids) == 4



import math
import re
from datetime import date, datetime
from typing import Optional

from dateutil.relativedelta import relativedelta
from pydantic import BaseModel

from recce.core import default_context
from recce.models import Check
from recce.tasks import Task
from recce.tasks.core import CheckValidator, TaskResultDiffer
from recce.tasks.query import QueryMixin

sql_datetime_types = [
    "DATE",
    "DATETIME",
    "TIMESTAMP",
    "TIME",
    "YEAR",  # Specific to MySQL/MariaDB
    "DATETIME2",
    "SMALLDATETIME",
    "DATETIMEOFFSET",  # Specific to SQL Server
    "INTERVAL",  # Common in PostgreSQL and Oracle
    "TIMESTAMPTZ",
    "TIMETZ",  # Specific to PostgreSQL
    "TIMESTAMP WITH TIME ZONE",
    "TIMESTAMP WITH LOCAL TIME ZONE",  # Oracle
    "TIMESTAMP_LTZ",
    "TIMESTAMP_NTZ",
    "TIMESTAMP_TZ",  # Specific to Snowflake
]

sql_integer_types = [
    "TINYINT",
    "SMALLINT",
    "MEDIUMINT",
    "INT",
    "INTEGER",
    "BIGINT",  # Common across most databases
    "INT2",
    "INT4",
    "INT8",  # PostgreSQL specific aliases
    "UNSIGNED BIG INT",  # SQLite specific
    "NUMBER",  # Oracle, can be used as an integer with precision and scale
    "NUMERIC",  # Generally available in many SQL databases, used with precision and scale
    "SMALLSERIAL",
    "SERIAL",
    "BIGSERIAL",  # PostgreSQL auto-increment types
    "IDENTITY",
    "SMALLIDENTITY",
    "BIGIDENTITY",  # SQL Server specific auto-increment types
    "BYTEINT",  # Specific to Snowflake, for storing very small integers
]

sql_not_supported_types = [
    "CHAR",
    "VARCHAR",
    "TINYTEXT",
    "TEXT",
    "MEDIUMTEXT",
    "LONGTEXT",
    "NCHAR",
    "NVARCHAR",
    "VARCHAR2",
    "NVARCHAR2",
    "CLOB",
    "NCLOB",
    "VARCHAR(MAX)",
    "XML",
    "JSON",
    "BOOLEAN",  # PostgreSQL, SQLite, and others with native boolean support
    "TINYINT(1)",  # MySQL/MariaDB uses TINYINT(1) to represent boolean values
    "BIT",  # SQL Server and others use BIT to represent boolean values, where 1 is true and 0 is false
    "NUMBER(1)",  # Oracle uses NUMBER(1) where 1 is true and 0 is false, as it does not have a native BOOLEAN type
    "BOOL",  # Snowflake and PostgreSQL also support BOOL as an alias for BOOLEAN
]

sql_not_supported_types_pattern = [
    r"^(CHAR|VARCHAR|NCHAR|NVARCHAR|VARCHAR2|NVARCHAR2)\(\d+\)$",  # String types with lengths
]


def _is_histogram_supported(column_type):
    if column_type.upper() in sql_not_supported_types:
        return False

    for pattern in sql_not_supported_types_pattern:
        if re.match(pattern, column_type.upper()):
            return False
    return True


def generate_histogram_sql_integer(node, column, min_value, max_value, num_bins=50):
    bin_size = math.ceil((max_value - min_value) / num_bins) or 1

    sql = f"""
    WITH value_ranges AS (
        SELECT
            {min_value} as min_value,
            {max_value} as max_value
    ),
    bin_parameters AS (
        SELECT
            min_value,
            max_value,
            {bin_size} AS bin_size
        FROM value_ranges
    ),
    binned_values AS (
        SELECT
            {column} as column_value,
            FLOOR(({column} - (SELECT min_value FROM bin_parameters)) / (SELECT bin_size FROM bin_parameters)) AS bin
        FROM {{{{ ref("{node}") }}}},
        bin_parameters
    ),
    bin_edges AS (
        SELECT
            bin,
            COUNT(*) AS count
        FROM binned_values, bin_parameters
        GROUP BY bin
        ORDER BY bin
    )

    SELECT bin, count FROM bin_edges
    """
    return sql, bin_size


def generate_histogram_sql_numeric(node, column, min_value, max_value, num_bins=50):
    bin_size = (max_value - min_value) / num_bins
    sql = f"""
        WITH value_ranges AS (
            SELECT
                {min_value} as min_value,
                {max_value} as max_value
        ),
        bin_parameters AS (
            SELECT
                min_value,
                max_value,
                {bin_size} AS bin_size
            FROM value_ranges
        ),
        binned_values AS (
            SELECT
                {column} as column_value,
                FLOOR(({column} - (SELECT min_value FROM bin_parameters)) / (SELECT bin_size FROM bin_parameters)) AS bin
            FROM {{{{ ref("{node}") }}}},
            bin_parameters
        ),
        bin_edges AS (
            SELECT
                bin,
                COUNT(*) AS count
            FROM binned_values, bin_parameters
            GROUP BY bin
            ORDER BY bin
        )

        SELECT bin, count FROM bin_edges
        """
    return sql, bin_size


class HistogramDiffParams(BaseModel):
    model: str
    column_name: str
    column_type: str
    num_bins: Optional[int] = 50


def query_numeric_histogram(task, node, column, column_type, min_value, max_value, num_bins=50):
    if column_type.upper() in sql_integer_types:
        if max_value - min_value < num_bins:
            num_bins = int(max_value - min_value + 1)
        histogram_sql, bin_size = generate_histogram_sql_integer(node, column, min_value, max_value, num_bins)
    else:
        histogram_sql, bin_size = generate_histogram_sql_numeric(node, column, min_value, max_value, num_bins)

    base = None
    try:
        base = task.execute_sql(histogram_sql, base=True)
    except Exception as e:
        print(e)
    finally:
        task.check_cancel()

    curr = None
    try:
        curr = task.execute_sql(histogram_sql, base=False)
    except Exception as e:
        print(e)
    finally:
        task.check_cancel()

    bin_edges = [None] * (num_bins + 1)
    labels = [""] * (num_bins + 1)

    base_result = {}
    curr_result = {}
    for i in range(num_bins + 1):
        val = int(min_value) + i * bin_size
        bin_edges[i] = val
        labels[i] = f"{val}-{val + bin_size}"

    if base is not None:
        counts = [0] * num_bins
        for row in base.rows:
            bin = row[0]
            count = row[1]
            if bin is not None:
                i = int(bin)
                if i < num_bins:
                    counts[i] = count
                else:
                    counts[num_bins - 1] += count
        base_result = {
            "counts": counts,
        }
    if curr is not None:
        counts = [0] * num_bins
        for row in curr.rows:
            bin = row[0]
            count = row[1]
            if bin is not None:
                i = int(bin)
                if i < num_bins:
                    counts[i] = count
                else:
                    counts[num_bins - 1] += count
        curr_result = {
            "counts": counts,
        }
    return base_result, curr_result, bin_edges, labels


def query_datetime_histogram(task, node, column, min_value, max_value):
    days_delta = (max_value - min_value).days
    print(max_value, min_value, days_delta)
    # _type = None
    if days_delta > 365 * 4:
        _type = "yearly"
        dmin = date(min_value.year, 1, 1)
        if max_value.year < 3000:
            dmax = date(max_value.year, 1, 1) + relativedelta(years=+1)
        else:
            dmax = date(3000, 1, 1)
        interval_years = math.ceil((dmax.year - dmin.year) / 50)
        interval = relativedelta(years=+interval_years)
        num_buckets = math.ceil((dmax.year - dmin.year) / interval.years)
        bin_edges = [dmin + relativedelta(year=i) for i in range(num_buckets + 1)]
        sql = f"""
        SELECT
            {{{{ date_trunc("year", "{column}") }}}} as year,
            COUNT(*) AS counts
        FROM {{{{ ref("{node}") }}}}
        WHERE {column} IS NOT NULL
        GROUP BY year
        ORDER BY year
        """
    elif days_delta > 60:
        _type = "monthly"
        interval = relativedelta(months=+1)
        dmin = date(min_value.year, min_value.month, 1)
        if max_value.year < 3000:
            dmax = date(max_value.year, max_value.month, 1) + interval
        else:
            dmax = date(3000, 1, 1)
        period = relativedelta(dmax, dmin)
        num_buckets = period.years * 12 + period.months
        bin_edges = [dmin + relativedelta(months=i) for i in range(num_buckets + 1)]
        sql = f"""
        SELECT
            {{{{ date_trunc("month", "{column}") }}}} as month,
            COUNT(*) AS counts
        FROM {{{{ ref("{node}") }}}}
        WHERE {column} IS NOT NULL
        GROUP BY month
        ORDER BY month
        """
    else:
        _type = "daily"
        interval = relativedelta(days=+1)
        dmin = date(min_value.year, min_value.month, min_value.day)
        if max_value.year < 3000:
            dmax = date(max_value.year, max_value.month, max_value.day) + interval
        else:
            dmax = date(3000, 1, 1)
        num_buckets = (dmax - dmin).days
        bin_edges = [dmin + relativedelta(day=i) for i in range(num_buckets + 1)]
        sql = f"""
        SELECT
            {{{{ date_trunc("day", "{column}") }}}} as day,
            COUNT(*) AS counts
        FROM {{{{ ref("{node}") }}}}
        WHERE {column} IS NOT NULL
        GROUP BY day
        ORDER BY day
        """

    base = None
    curr = None
    try:
        base = task.execute_sql(sql, base=True)
    except Exception as e:
        print(e)
    finally:
        task.check_cancel()
    try:
        curr = task.execute_sql(sql, base=False)
    except Exception as e:
        print(e)
    finally:
        task.check_cancel()

    base_counts = [0] * num_buckets
    print(_type)
    for d, v in base.rows:
        i = bin_edges.index(d.date()) if isinstance(d, datetime) else bin_edges.index(d)
        base_counts[i] = v
    curr_counts = [0] * num_buckets
    for d, v in curr.rows:
        i = bin_edges.index(d.date()) if isinstance(d, datetime) else bin_edges.index(d)
        curr_counts[i] = v
    base_result = {
        "counts": base_counts,
    }
    curr_result = {
        "counts": curr_counts,
    }

    return base_result, curr_result, bin_edges


class HistogramDiffTask(Task, QueryMixin):
    def __init__(self, params):
        super().__init__()
        self.params = HistogramDiffParams(**params)
        self.connection = None

    def execute(self):
        from recce.adapter.dbt_adapter import DbtAdapter

        result = {}

        dbt_adapter: DbtAdapter = default_context().adapter
        node = self.params.model
        column = self.params.column_name
        num_bins = self.params.num_bins or 50
        column_type = self.params.column_type

        if _is_histogram_supported(column_type) is False:
            raise ValueError(f"Column type {column_type} is not supported for histogram analysis")

        with dbt_adapter.connection_named("query"):
            self.connection = dbt_adapter.get_thread_connection()
            min_max_sql = f"""
                SELECT
                    MIN({column}) as min,
                    MAX({column}) as max,
                    COUNT({column}) as total
                FROM {{{{ ref("{node}") }}}}
                """
            # Get the mix/max values from both the base and current environments

            min_max_base = self.execute_sql(min_max_sql, base=True)
            min_max_curr = self.execute_sql(min_max_sql, base=False)

            def get_min_max(fn, base, curr):
                if base is None and curr is None:
                    return None
                if base is None:
                    return curr
                if curr is None:
                    return base
                return fn(base, curr)

            min_value = get_min_max(min, min_max_base[0][0], min_max_curr[0][0])
            max_value = get_min_max(max, min_max_base[0][1], min_max_curr[0][1])
            base_total = min_max_base[0][2]
            curr_total = min_max_curr[0][2]

            # Get histogram data from both the base and current environments
            labels = None
            if min_value is None or max_value is None:
                base_result = {
                    "counts": [],
                }
                current_result = {
                    "counts": [],
                }
                bin_edges = []
                labels = []
            elif column_type.upper() in sql_datetime_types:
                base_result, current_result, bin_edges = query_datetime_histogram(
                    self, node, column, min_value, max_value
                )
            else:
                base_result, current_result, bin_edges, labels = query_numeric_histogram(
                    self, node, column, column_type, min_value, max_value, num_bins
                )
            if base_result:
                base_result["total"] = base_total
            if current_result:
                current_result["total"] = curr_total
            result["base"] = base_result
            result["current"] = current_result
            result["min"] = min_value
            result["max"] = max_value
            result["bin_edges"] = bin_edges
            result["labels"] = labels
        return result

    def cancel(self):
        super().cancel()
        if self.connection:
            self.close_connection(self.connection)


class HistogramDiffTaskResultDiffer(TaskResultDiffer):
    def _check_result_changed_fn(self, result):
        return TaskResultDiffer.diff(result["base"], result["current"])


class HistogramDiffCheckValidator(CheckValidator):

    def validate_check(self, check: Check):
        try:
            HistogramDiffParams(**check.params)
        except Exception as e:
            raise ValueError(f"Invalid check: {str(e)}")



"""Define the type to serialize/de-serialize the state of the recce instance."""

import json
import logging
from datetime import datetime
from typing import Dict, List, Optional

from pydantic import BaseModel, Field

from recce import get_version
from recce.config import console
from recce.exceptions import RecceException
from recce.git import current_branch
from recce.models.types import Check, Run
from recce.pull_request import PullRequestInfo
from recce.util.io import SupportedFileTypes, file_io_factory
from recce.util.pydantic_model import pydantic_model_dump, pydantic_model_json_dump

logger = logging.getLogger("uvicorn")


class GitRepoInfo(BaseModel):
    branch: Optional[str] = None

    @staticmethod
    def from_current_repository() -> Optional["GitRepoInfo"]:
        branch = current_branch()
        if branch is None:
            return None

        return GitRepoInfo(branch=branch)

    def to_dict(self):
        return pydantic_model_dump(self)


class RecceStateMetadata(BaseModel):
    dbt_version: str = "0"
    schema_version: str = "v0"
    recce_version: str = Field(default_factory=lambda: get_version())
    generated_at: str = Field(default_factory=lambda: datetime.utcnow().strftime("%Y-%m-%dT%H:%M:%SZ"))


class ArtifactsRoot(BaseModel):
    """
    Root of the artifacts.

    base: artifacts of the base env. key is file name, value is dict
    current: artifacts of the current env. key is file name, value is dict
    """

    base: Dict[str, Optional[dict]] = {}
    current: Dict[str, Optional[dict]] = {}


class RecceState(BaseModel):
    metadata: Optional[RecceStateMetadata] = None
    runs: Optional[List[Run]] = Field(default_factory=list)
    checks: Optional[List[Check]] = Field(default_factory=list)
    artifacts: ArtifactsRoot = ArtifactsRoot(base={}, current={})
    git: Optional[GitRepoInfo] = None
    pull_request: Optional[PullRequestInfo] = None

    @staticmethod
    def from_json(json_content: str):
        dict_data = json.loads(json_content)
        state = RecceState(**dict_data)
        metadata = state.metadata

        from recce.adapter.dbt_adapter import DbtVersion

        dbt_version = DbtVersion()
        if metadata:
            if metadata.schema_version is None:
                pass
            if metadata.schema_version == "v0":
                pass
            else:
                raise RecceException(f"Unsupported state file version: {metadata.schema_version}")
            if metadata.dbt_version != dbt_version:
                console.print(
                    f"[[yellow]WARN[/yellow]] dbt version mismatch. Local dbt adapter: {dbt_version} vs. Generated Manifest dbt adapter: {metadata.dbt_version}"
                )
                console.print("[[yellow]WARN[/yellow]] Version mismatch can lead to issues when generating queries")
        return state

    @staticmethod
    def from_file(file_path: str, file_type: SupportedFileTypes = SupportedFileTypes.FILE):
        """
        Load the state from a recce state file.
        """
        from pathlib import Path

        logger.debug(f"Load state file from: '{file_path}'")
        if not Path(file_path).is_file():
            return None

        io = file_io_factory(file_type)
        json_content = io.read(file_path)
        return RecceState.from_json(json_content)

    def to_json(self):
        return pydantic_model_json_dump(self)

    def to_file(self, file_path: str, file_type: SupportedFileTypes = SupportedFileTypes.FILE):

        json_data = self.to_json()
        io = file_io_factory(file_type)

        io.write(file_path, json_data)
        return f"The state file is stored at '{file_path}'"

    def _merge_run(self, run: Run):
        for r in self.runs:
            if r.run_id == run.run_id:
                break
        else:
            self.runs.append(run)

    def _merge_check(self, check: Check):
        for c in self.checks:
            if c.check_id == check.check_id:
                c.merge(check)
                break
        else:
            self.checks.append(check)

    def _merge_artifacts(self, artifacts: ArtifactsRoot):
        self.artifacts.merge(artifacts)
